from dataclasses import dataclass
from typing import Tuple, Dict, Callable

from netaddr import IPAddress

from cyst.api.environment.configuration import EnvironmentConfiguration
from cyst.api.environment.interpreter import ActionInterpreter, ActionInterpreterDescription
from cyst.api.environment.message import Request, Response, Status, StatusOrigin, StatusValue, StatusDetail
from cyst.api.environment.messaging import EnvironmentMessaging
from cyst.api.environment.policy import EnvironmentPolicy
from cyst.api.environment.resources import EnvironmentResources
from cyst.api.logic.access import AuthenticationTarget, Authorization
from cyst.api.logic.action import ActionDescription, ActionToken, ActionParameterType, ActionParameter
from cyst.api.network.node import Node


@dataclass
class AuthenticationTracker:
    src_ip: str
    src_service: str
    dst_ip: str
    dst_service: str
    identity: str
    step: int
    time: int
    attempts: int


class METAInterpreter(ActionInterpreter):
    def __init__(self, configuration: EnvironmentConfiguration, resources: EnvironmentResources,
                 policy: EnvironmentPolicy, messaging: EnvironmentMessaging) -> None:

        self._configuration = configuration
        self._action_store = resources.action_store
        self._exploit_store = resources.exploit_store
        self._policy = policy
        self._messaging = messaging
        self._authentications: Dict[Tuple[str, str, str, str, str], AuthenticationTracker] = {}

        self._action_store.add(ActionDescription("meta:inspect:node",
                                                 "Discovery of hosts in a network. Equivalent to ping scanning.",
                                                 [],
                                                 [(ActionToken.SESSION, ActionToken.NONE)]))

        self._action_store.add(ActionDescription("meta:authenticate",
                                                 "Authentication against a service.",
                                                 [ActionParameter(ActionParameterType.TOKEN, "auth_token", configuration.action.create_action_parameter_domain_any())],
                                                 [(ActionToken.NONE, ActionToken.NONE)])) # Tokens are wrong, I know that

    def evaluate(self, message: Request, node: Node) -> Tuple[int, Response]:
        if not message.action:
            raise ValueError("Action not provided")

        action_name = "_".join(message.action.fragments)
        fn: Callable[[Request, Node], Tuple[int, Response]] = getattr(self, "process_" + action_name, self.process_default)
        return fn(message, node)

    def process_default(self, message: Request, node: Node) -> Tuple[int, Response]:
        print("Could not evaluate message. Tag in `meta` namespace unknown. " + str(message))
        return 0, self._messaging.create_response(message, status=Status(StatusOrigin.SYSTEM, StatusValue.ERROR), session=message.session)

    def process_inspect_node(self, message: Request, node: Node) -> Tuple[int, Response]:
        error = ""
        if message.src_ip != IPAddress("127.0.0.1"):  # Local IP has a free pass
            if not message.session:
                error = "Session not provided"
            elif message.session.end[0] not in [x.ip for x in node.interfaces]:
                error = "Session does not end in required node"

        if error:
            return 0, self._messaging.create_response(message, Status(StatusOrigin.NETWORK, StatusValue.FAILURE), error, session=message.session)

        return 1, self._messaging.create_response(message, Status(StatusOrigin.NODE, StatusValue.SUCCESS),
                                                  node, session=message.session, auth=message.auth)

    def process_authenticate(self, message: Request, node: Node) -> Tuple[int, Response]:
        # To authenticate, an actor has to go through all the phases in the authentication scheme.
        # After each factor in authentication scheme is passed, the actor is given an authorization, which enables them
        # to attempt a next step. If it is the last step, the full authorization for this authentication scheme is
        # given.

        src_node = str(message.src_ip)
        src_service = message.src_service
        dst_node = str(message.dst_ip)
        dst_service = message.dst_service

        # First of all, check if the message contains authentication token
        token_found = False
        token = None
        t = message.action.parameters.get("auth_token", None)
        if t:
            if not isinstance(t, ActionParameter):
                raise RuntimeError("Wrong object type supplied as action parameter. Type: " + str(type(t)))

            if t.type == ActionParameterType.TOKEN and t.value:
                token_found = True
                token = t.value

        if not token_found:
            return 0, self._messaging.create_response(message, Status(StatusOrigin.SERVICE, StatusValue.FAILURE, StatusDetail.AUTHENTICATION_NOT_PROVIDED),
                                                      "No auth token provided", session=message.session, auth=message.auth)

        #  check if node has the target service
        s = node.services.get(dst_service)
        if s is None:
            return 0, self._messaging.create_response(message, Status(StatusOrigin.SERVICE, StatusValue.FAILURE,
                                                                      StatusDetail.AUTHENTICATION_NOT_PROVIDED),
                                                      "Service does not exist on this node", session=message.session,
                                                      auth=message.auth)


        #  evaluate the token
        result = self._configuration.access.evaluate_token_for_service(s, token, node, message.dst_ip)

        if result is None:
            return 0, self._messaging.create_response(message, Status(StatusOrigin.SERVICE, StatusValue.FAILURE,
                                                                      StatusDetail.AUTHENTICATION_NOT_APPLICABLE),
                                                      "Token invalid for this service", session=message.session,
                                                      auth=message.auth)

        if isinstance(result, AuthenticationTarget):
            return 0, self._messaging.create_response(message, Status(StatusOrigin.SERVICE, StatusValue.FAILURE,
                                                                      StatusDetail.AUTHENTICATION_NEXT),
                                                      "Continue with next factor", session=message.session,
                                                      auth=result)

        if isinstance(result, Authorization):
            return 0, self._messaging.create_response(message, Status(StatusOrigin.SERVICE, StatusValue.SUCCESS),
                                                      "Authorized", session=message.session,
                                                      auth=result)




def create_meta_interpreter(configuration: EnvironmentConfiguration, resources: EnvironmentResources,
                            policy: EnvironmentPolicy, messaging: EnvironmentMessaging) -> ActionInterpreter:
    interpreter = METAInterpreter(configuration, resources, policy, messaging)
    return interpreter


action_interpreter_description = ActionInterpreterDescription(
    "meta",
    "Interpreter for auxiliary actions needed to supplement",
    create_meta_interpreter
)


import argparse
import copy
import logging
import os
import sys

from cyst.api.environment.interpreter import ActionInterpreter

if sys.version_info < (3, 10):
    from importlib_metadata import entry_points
else:
    from importlib.metadata import entry_points

from heapq import heappush, heappop
from itertools import product
from time import localtime
from typing import Tuple, List, Union, Optional, Any, Dict

from cyst.api.configuration.configuration import ConfigItem
from cyst.api.environment.environment import Environment
from cyst.api.environment.control import EnvironmentState, EnvironmentControl
from cyst.api.environment.configuration import EnvironmentConfiguration, GeneralConfiguration, NodeConfiguration, \
    ServiceConfiguration, NetworkConfiguration, ExploitConfiguration, AccessConfiguration, ActionConfiguration
from cyst.api.environment.messaging import EnvironmentMessaging
from cyst.api.environment.metadata_provider import MetadataProvider
from cyst.api.environment.policy import EnvironmentPolicy
from cyst.api.environment.resources import EnvironmentResources
from cyst.api.environment.message import Message, MessageType, Request, StatusValue, StatusOrigin, Status, StatusDetail, Timeout
from cyst.api.logic.access import AuthenticationToken
from cyst.api.network.node import Node
from cyst.api.network.session import Session
from cyst.api.network.firewall import FirewallRule, FirewallPolicy
from cyst.api.host.service import Service, PassiveService, ActiveService, ServiceState
from cyst.api.configuration.configuration import ConfigItem
from cyst.api.utils.counter import Counter

from cyst.core.environment.configuration_access import AccessConfigurationImpl
from cyst.core.environment.configuration_action import ActionConfigurationImpl
from cyst.core.environment.configuration_exploit import ExploitConfigurationImpl
from cyst.core.environment.configuration_general import GeneralConfigurationImpl, RuntimeConfiguration
from cyst.core.environment.configuration_network import NetworkConfigurationImpl
from cyst.core.environment.configuration_node import NodeConfigurationImpl
from cyst.core.environment.configuration_service import ServiceConfigurationImpl
from cyst.core.environment.environment_control import EnvironmentControlImpl
from cyst.core.environment.environment_messaging import EnvironmentMessagingImpl
from cyst.core.environment.environment_resources import EnvironmentResourcesImpl
from cyst.core.environment.data_store import DataStore
from cyst.core.environment.message import MessageImpl, RequestImpl, ResponseImpl, TimeoutImpl
from cyst.core.environment.proxy import EnvironmentProxy
from cyst.core.environment.stores import ServiceStoreImpl
from cyst.core.host.service import ServiceImpl
from cyst.core.logic.access import AuthenticationTokenImpl
from cyst.core.logic.policy import Policy
from cyst.core.network.elements import Endpoint, InterfaceImpl, Hop
from cyst.core.network.firewall import service_description as firewall_service_description
from cyst.core.network.network import Network
from cyst.core.network.node import NodeImpl
from cyst.core.network.router import Router
from cyst.core.network.session import SessionImpl


# Environment is unlike other core implementation given an underscore-prefixed name to let python complain about
# it being private if instantiated otherwise than via the create_environment()
class _Environment(Environment, EnvironmentConfiguration):

    def __init__(self) -> None:
        self._time = 0
        self._start_time = localtime()
        self._tasks: List[Tuple[int, int, MessageImpl]] = []
        self._pause = False
        self._terminate = False
        self._initialized = False
        self._state = EnvironmentState.INIT

        self._run_id = ""

        self._pause_on_request: List[str] = []
        self._pause_on_response: List[str] = []

        # Interface implementations
        self._environment_control = EnvironmentControlImpl(self)
        self._environment_messaging = EnvironmentMessagingImpl(self)
        self._environment_resources = EnvironmentResourcesImpl(self)

        self._service_store = ServiceStoreImpl(self.messaging, self.resources)

        self._interpreters: Dict[str, ActionInterpreter] = {}
        # TODO currently, there can be only on metadata provider for one namespace
        self._metadata_providers: Dict[str, MetadataProvider] = {}

        self._policy = Policy(self)

        self._sessions_to_add: List[Tuple[ str, List[Union[str, Node]],Optional[str] ,Optional[str], Optional[Session], bool]] = []

        self._access_configuration = AccessConfigurationImpl(self)
        self._action_configuration = ActionConfigurationImpl(self)
        self._exploit_configuration = ExploitConfigurationImpl(self)
        self._general_configuration = GeneralConfigurationImpl(self)
        self._network_configuration = NetworkConfigurationImpl(self)
        self._node_configuration = NodeConfigurationImpl(self)
        self._service_configuration = ServiceConfigurationImpl(self)
        self._runtime_configuration = RuntimeConfiguration()

        self._configure_runtime()
        self._register_services()
        self._register_actions()
        self._register_metadata_providers()

        self._network = Network(self._general_configuration)

        self._data_store = DataStore(self._runtime_configuration.data_backend,
                                     self._runtime_configuration.data_backend_params)

        # Logs
        self._message_log = logging.getLogger("messaging")

    def __getstate__(self) -> dict:
        return {
            # Simple values
            "_time": self._time,
            "_start_time": self._start_time,
            "_pause": self._pause,
            "_terminate": self._terminate,
            "_initialized": self._initialized,
            "_state": self._state,
            "_run_id": self._run_id,

            # Arrays
            "_pause_on_response": self._pause_on_response,
            "_pause_on_request": self._pause_on_request,

            # Simple objects
            "_runtime_configuration": self._runtime_configuration,

            # Complex beasts
            "_service_store": self._service_store,
            "_environment_resources": self._environment_resources,
            "_metadata_providers": self._metadata_providers,
            "_network": self._network,
            "_general_configuration": self._general_configuration

            # Ignored members
            # Policy - is reinitialized, no need to serialize
            # DataStore - stays the same across serializations
            # Log - stays the same across serializations
            # All interface implementations excluding the general configuration and environment resources
        }

    def __setstate__(self, state: dict) -> None:
        self._time = state["_time"]
        self._start_time = state["_start_time"]
        self._pause = state["_pause"]
        self._terminate = state["_terminate"]
        self._initialized = state["_initialized"]
        self._state = state["_state"]
        self._run_id = state["_run_id"]

        self._pause_on_response = state["_pause_on_response"]
        self._pause_on_request = state["_pause_on_request"]

        self._runtime_configuration = state["_runtime_configuration"]

        self._service_store = state["_service_store"]
        self._environment_resources = state["_environment_resources"]
        self._metadata_providers = state["_metadata_providers"]
        self._network = state["_network"]
        self._general_configuration = state["_general_configuration"]

        # Members reconstructed on the fly
        self._policy = Policy(self)

        self._environment_control = EnvironmentControlImpl(self)
        self._environment_messaging = EnvironmentMessagingImpl(self)

        self._access_configuration = AccessConfigurationImpl(self)
        self._action_configuration = ActionConfigurationImpl(self)
        self._exploit_configuration = ExploitConfigurationImpl(self)
        self._network_configuration = NetworkConfigurationImpl(self)
        self._node_configuration = NodeConfigurationImpl(self)
        self._service_configuration = ServiceConfigurationImpl(self)

    # Replace the environment with the state of another environment. This is used for deserialization. It is explicit to
    # avoid replacing of ephemeral stuff, such as data store connections or whatnot
    def _replace(self, env: "_Environment"):
        self._time = env._time
        self._start_time = env._start_time
        self._pause = env._pause
        self._terminate = env._terminate
        self._initialized = env._initialized
        self._state = env._state
        self._run_id = env._run_id

        self._pause_on_response = env._pause_on_response
        self._pause_on_request = env._pause_on_request

        self._runtime_configuration = env._runtime_configuration

        self._service_store = env._service_store
        self._environment_resources = env._environment_resources
        self._metadata_providers = env._metadata_providers
        self._network = env._network
        self._general_configuration = env._general_configuration

        self._policy = env._policy

    # Runtime parameters can be passed via command-line, configuration file, or through environment variables
    # In case of multiple definitions of one parameter, the order is, from the most important to least:
    #                                                            command line, configuration file, environment variables
    def _configure_runtime(self) -> None:
        # Environment
        data_backend = os.environ.get('CYST_DATA_BACKEND')
        data_backend_params: Dict[str, str] = dict()
        if data_backend:
            data_backend_params_serialized = os.environ.get('CYST_DATA_BACKEND_PARAMS')
            # we expect parameters to be given in the form "param1_name","param1_value","param2_name","param2_value",...
            if data_backend_params_serialized:
                data_backend_params = dict(tuple(x) for x in data_backend_params_serialized.split(',').islice(2))
        run_id = os.environ.get('CYST_RUN_ID')
        config_id = os.environ.get('CYST_CONFIG_ID')
        config_filename = os.environ.get('CYST_CONFIG_FILENAME')

        # Command line (only parsing)
        cmdline_parser = argparse.ArgumentParser(description="CYST runtime configuration")

        cmdline_parser.add_argument("-c", "--config_file", type=str,
                                    help="Path to a file storing the configuration. Commandline overrides the items in configuration file.")
        cmdline_parser.add_argument("-b", "--data_backend", type=str,
                                    help="The type of a backend to use. Currently supported are: REDIS")
        cmdline_parser.add_argument("-p", "--data_backend_parameter", action="append", nargs=2, type=str,
                                    metavar=('NAME', 'VALUE'), help="Parameters to be passed to data backend.")
        cmdline_parser.add_argument("-r", "--run_id", type=str,
                                    help="A unique identifier of a simulation run. If not specified, a UUID will be generated instead.")
        cmdline_parser.add_argument("-i", "--config_id", type=str,
                                    help="A unique identifier of simulation run configuration, which can be obtained from the data store.")

        args, _ = cmdline_parser.parse_known_args()
        if args.config_file:
            config_filename = args.config_file

        # --------------------------------------------------------------------------------------------------------------
        # Config file TODO
        if config_filename:
            pass
        # --------------------------------------------------------------------------------------------------------------

        # Command line argument override
        if args.data_backend:
            data_backend = args.data_backend

        if args.data_backend_parameter:
            # Convert from list of lists into a list of tuples
            data_backend_params = dict(tuple(x) for x in args.data_backend_parameter) #MYPY: typehinting lambda not really possible this way, better to ignore?

        if args.run_id:
            run_id = args.run_id

        if args.config_id:
            config_id = args.config_id

        # --------------------------------------------------------------------------------------------------------------
        if data_backend:  # Fuck, I miss oneliners
            self._runtime_configuration.data_backend = data_backend
        if data_backend_params:
            self._runtime_configuration.data_backend_params = data_backend_params
        if config_filename:
            self._runtime_configuration.config_filename = config_filename
        if run_id:
            self._runtime_configuration.run_id = run_id
        if config_id:
            self._runtime_configuration.config_id = config_id

    # ------------------------------------------------------------------------------------------------------------------
    # Environment.
    @property
    def configuration(self) -> EnvironmentConfiguration:
        return self

    @property
    def control(self) -> EnvironmentControl:
        return self._environment_control

    @property
    def messaging(self) -> EnvironmentMessaging:
        return self._environment_messaging

    @property
    def resources(self) -> EnvironmentResources:
        return self._environment_resources

    @property
    def policy(self) -> EnvironmentPolicy:
        return self._policy

    def configure(self, *config_item: ConfigItem) -> Environment:
        return self._general_configuration.configure(*[copy.deepcopy(x) for x in config_item])

    # ------------------------------------------------------------------------------------------------------------------
    # EnvironmentConfiguration
    # ------------------------------------------------------------------------------------------------------------------
    # Just point on itself
    @property
    def general(self) -> GeneralConfiguration:
        return self._general_configuration

    @property
    def node(self) -> NodeConfiguration:
        return self._node_configuration

    @property
    def service(self) -> ServiceConfiguration:
        return self._service_configuration

    @property
    def network(self) -> NetworkConfiguration:
        return self._network_configuration

    @property
    def exploit(self) -> ExploitConfiguration:
        return self._exploit_configuration

    @property
    def action(self) -> ActionConfiguration:
        return self._action_configuration

    @property
    def access(self) -> AccessConfiguration:
        return self._access_configuration

    # ------------------------------------------------------------------------------------------------------------------
    # Internal functions
    @property
    def _get_network(self) -> Network:
        return self._network

    # When creating sessions from nodes, there are two options - either nodes are connected directly, or they
    # go through a router. So correct hops are evaluated either in N-R*-N form or N-N
    # TODO: If one direction fails, session should try constructing itself in reverse order and then restructure hops
    #       so that the origin is always at the first waypoint.
    def _create_session(self, owner: str, waypoints: List[Union[str, Node]], src_service: Optional[str],
                        dst_service: Optional[str], parent: Optional[Session], reverse: bool) -> Session:
        path: List[Hop] = []
        source: NodeImpl
        session_reversed = False

        if len(waypoints) < 2:
            raise ValueError("The session path needs at least two ids")

        session_constructed = True
        for direction in ("forward", "reverse"):

            if direction == "reverse":
                if not session_constructed:
                    path.clear()
                    waypoints.reverse()
                    session_reversed = True
                    session_constructed = True
                else:
                    break

            i = 0
            while i < len(waypoints) - 1:
                # There was an error in partial session construction
                if not session_constructed:
                    break

                node0 = None
                node1 = None
                node2 = None

                def get_node_from_waypoint(self, i: int) -> Node:
                    if isinstance(waypoints[i], str):
                        node = self._network.get_node_by_id(waypoints[i])
                    else:
                        node = waypoints[i]
                    return node

                # Get the nodes
                node0 = get_node_from_waypoint(self, i)
                node1 = get_node_from_waypoint(self, i + 1)

                routers = []
                # N-R*-N
                if node1.type == "Router":
                    router = Router.cast_from(node1)

                    routers.append(router)
                    node2 = get_node_from_waypoint(self, i + len(routers) + 1)

                    while node2.type == "Router":
                        routers.append(Router.cast_from(node2))
                        node2 = get_node_from_waypoint(self, i + len(routers) + 1)


                    path_candidate: List[Hop] = []
                    for elements in product(node0.interfaces, node2.interfaces):
                        node0_iface = InterfaceImpl.cast_from(elements[0])
                        node2_iface = InterfaceImpl.cast_from(elements[1])

                        path_candidate.clear()

                        # Check if the next router is connected to the first node
                        if node0_iface.endpoint.id != routers[0].id:
                            continue

                        # It is, so it's a first hop
                        path_candidate.append(
                            Hop(Endpoint(NodeImpl.cast_from(node0).id, node0_iface.index, node0_iface.ip),
                                node0_iface.endpoint))

                        # Check for every router if it routes the source and destination
                        for j, r in enumerate(routers):
                            # Find if there is a forward port
                            # Ports are returned in order of priority: local IPs, remote IPs sorted by specificity (CIDR)
                            port = r.routes(node0_iface.ip, node2_iface.ip, "*")

                            # No suitable port found, try again
                            if not port:
                                break

                            path_candidate.append(Hop(Endpoint(r.id, port.index, port.ip), port.endpoint))

                        if len(path_candidate) == len(routers) + 1:
                            path.extend(path_candidate)
                            break

                    i += len(routers) + 1

                    if len(path) < i:
                        session_constructed = False
                        break
                        # raise RuntimeError("Could not find connection between {} and {} to establish a session".format(NodeImpl.cast_from(node0).id, NodeImpl.cast_from(node2).id))
                else:
                    # N-N
                    for iface in node0.interfaces:
                        node0_iface = InterfaceImpl.cast_from(iface)

                        if node0_iface.endpoint.id == NodeImpl.cast_from(node1).id:
                            path.append(Hop(Endpoint(NodeImpl.cast_from(node0).id, node0_iface.index, node0_iface.ip),
                                            node0_iface.endpoint))
                            break

                    i += 1
                    if len(path) < i:
                        session_constructed = False
                        break
                        # raise RuntimeError("Could not find connection between {} and {} to establish a session".format(NodeImpl.cast_from(node0).id, NodeImpl.cast_from(node1).id))

        if not session_constructed:
            # Sessions are always tried to be constructed in both directions, so we need to reverse the waypoints again
            waypoints.reverse()
            raise RuntimeError(
                "Could not find connection between the following waypoints to establish a session".format(waypoints)) #MYPY: Missing the parameter in string

        # If the session was constructed from the end to front, we need to reverse the path
        if session_reversed:
            path.reverse()
            for i in range(0, len(path)):
                path[i] = path[i].swap()

        return SessionImpl(owner, parent, path, src_service, dst_service, self._network) #MYPY: Services can be None, they are optional

    def _process_passive(self, message: Request, node: Node):
        time = 0
        response = None

        message = RequestImpl.cast_from(message)

        # TODO: auto-authentication here, maybe??
        # cyst namespace is currently disabled from auto authentication
        if message.auth and isinstance(message.auth, AuthenticationToken):
            if not AuthenticationTokenImpl.is_local_instance(message.auth):
                return time, self.messaging.create_response(message, Status(StatusOrigin.SERVICE,
                                                                            StatusValue.FAILURE,
                                                                            StatusDetail.AUTHENTICATION_NOT_APPLICABLE),
                                                            "Auto-authentication does not work with non-local tokens",
                                                            session=message.session, auth=message.auth) #MYPY: AuthenticationToken is not valid type here. Depends, if they were just forgotten in annotation or not
            # TODO: assess if locality check makes sense
            original_action = message.action
            auth_action = self._environment_resources.action_store.get("meta:authenticate").copy()
            auth_action.parameters["auth_token"].value = message.auth
            message.action = auth_action  # swap to authentication
            auth_time, auth_response = self._interpreters["meta"].evaluate(message, node)

            if auth_response.status.value == StatusValue.FAILURE:  # not authorized
                return auth_time, auth_response

            message.auth = auth_response.auth  # if authentication successful, swap auth token for authorization
            message.action = original_action  # swap back original action
        #  and continue to action

        # TODO: In light of tags abandonment, how to deal with splitting id into namespace and action name
        if message.action.namespace in self._interpreters:
            time, response = self._interpreters[message.action.namespace].evaluate(message, node)

        return time, response

    def _send_message(self, message: MessageImpl) -> None:
        message_type = "request" if isinstance(message, Request) else "response"

        # shortcut for wakeup messages
        if message.type == MessageType.TIMEOUT:
            self._network.get_node_by_id(message.origin.id).process_message(message)  #MYPY: Node returned by get_node can be None
            return

        # Store it into the history
        self._data_store.set(self._run_id, message, Message)

        # Move message to a next hop
        message.hop()
        current_node: NodeImpl = self._network.get_node_by_id(message.current.id) #MYPY: Get node can return None

        connection = self.configuration.network.get_connections(current_node, message.current.port)[0]
        delay, result = connection.evaluate(message)
        if delay < 0:
            # TODO: message dropped, what to do? Maybe send early without processing
            pass

        message = MessageImpl.cast_from(result)
        processing_time = max(0, delay)

        # HACK: Because we want to enable actions to be able to target routers, we need to bypass the router processing
        #       if the message is at the end of its journey
        last_hop = message.dst_ip == message.current.ip #MYPY: current can return None

        if not last_hop and current_node.type == "Router":
            result, delay = current_node.process_message(message) #MYPY: This only returns one int, will crash
            processing_time += delay
            if result:
                heappush(self._tasks, (self._time + processing_time, Counter().get("msg"), message))

            return

        # Message has a session
        if message.session:
            local_processing = False
            # Message still in session, pass it along
            if message.in_session:
                message.set_next_hop()
                heappush(self._tasks, (self._time + processing_time, Counter().get("msg"), message))
                return
            # The session ends in the current node
            elif message.session.endpoint.id == current_node.id or message.session.startpoint.id == current_node.id:  #MYPY: here on multiple line, session only has an end and start, not endpoint and startpoint
                # TODO bi-directional session complicate the situation soooo much
                end_port = None
                if message.session.endpoint.id == current_node.id:
                    end_port = message.session.endpoint.port
                elif message.session.startpoint.id == current_node.id:
                    end_port = message.session.startpoint.port

                # Check if the node is the final destination
                for iface in current_node.interfaces:
                    if iface.index == end_port and iface.ip == message.dst_ip: #MYPY: Interface does not have index
                        local_processing = True
                        break
                # It is not, this means the node was only a proxy to some other target
                if not local_processing:
                    # Find a way to nearest switch
                    gateway, port = current_node.gateway(message.dst_ip) #MYPY: If this returns None, there is only one value and it will crash on unpacking it
                    # ##################
                    dest_node_endpoint = current_node.interfaces[port].endpoint #MYPY: end vs endpoint
                    dest_node = self._network.get_node_by_id(dest_node_endpoint.id)
                    dest_node_ip = dest_node.interfaces[dest_node_endpoint.port].ip #MYPY: dest_node can be None
                    message.set_next_hop(Endpoint(current_node.id, port, current_node.interfaces[port].ip),
                                         Endpoint(dest_node_endpoint.id, dest_node_endpoint.port, dest_node_ip))
                    # ##################
                    self._message_log.debug(
                        f"Proxying {message_type} to {message.dst_ip} via {message.next_hop.id} on a node {current_node.id}")
                    heappush(self._tasks, (self._time + processing_time, Counter().get("msg"), message))
                    return

        # Message has to be processed locally
        self._message_log.debug(f"Processing {message_type} on a node {current_node.id}. {message}")

        # Before a message reaches to services within, it is evaluated by all traffic processors
        # While they are returning true, everything is ok. Once they return false, the message processing stops
        # Traffic processors are free to send any reply as they see fit
        # TODO: Firewall does not return a response and currently we want it in some instances to return it and in
        #       some instances we don't. This is not a good situation.
        for processor in current_node.traffic_processors:
            result, delay = processor.process_message(message)
            if not result:
                return

        # Service is requested
        response = None
        if message.dst_service:
            # Check if the requested service exists on the current node
            if message.dst_service not in current_node.services:
                # There is a theoretical chance for response not finding dst service for responses, if e.g. attacker
                # shut down the service after firing request and before receiving the response. In such case the
                # error is silently dropped
                if message_type == "response":
                    return

                processing_time += 1
                response = ResponseImpl(message, Status(StatusOrigin.NODE, StatusValue.ERROR),
                                        "Nonexistent service {} at node {}".format(message.dst_service, message.dst_ip),
                                        session=message.session, auth=message.auth)
                self._environment_messaging.send_message(response, processing_time)

            # Service exists and it is passive
            elif current_node.services[message.dst_service].passive:  #MYPY: passive vs .passive_service
                # Passive services just discard the responses and only process the requests
                if message_type == "response":
                    return

                if current_node.services[message.dst_service].passive_service.state != ServiceState.RUNNING:
                    response = ResponseImpl(message, Status(StatusOrigin.NODE, StatusValue.ERROR),
                                            "Service {} at node {} is not running".format(message.dst_service, message.dst_ip),
                                            session=message.session, auth=message.auth)
                    self._environment_messaging.send_message(response, processing_time)
                else:
                    delay, response = self._process_passive(message, current_node)
                    processing_time += delay
                    if response.status.origin == StatusOrigin.SYSTEM and response.status.value == StatusValue.ERROR:
                        print("Could not process the request, unknown semantics.")
                    else:
                        self._environment_messaging.send_message(response, processing_time)
            # Service exists and it is active
            else:
                # An active service does not necessarily produce Responses, so we should just move time
                # somehow and be done with it.
                # TODO How to move time?
                result, delay = current_node.services[message.dst_service].active_service.process_message(message)

                if message_type == "response" and current_node.id + "." + message.dst_service in self._pause_on_response:
                    self._pause = True

        # If no service is specified, it is a message to a node, but still, it is processed as a request for
        # passive service and processed with the interpreter
        # No service is specified
        else:
            # If there is response arriving without destination service, just drop it
            if message_type == "response":
                return

            # If it is a request, then it is processed as a request for passive service and processed with the interpreter
            delay, response = self._process_passive(message, current_node) #MYPY: messageimpl vs request
            processing_time += delay
            if response.status.origin == StatusOrigin.SYSTEM and response.status.value == StatusValue.ERROR: #MYPY: same as above, response None?
                print("Could not process the request, unknown semantics.")
            else:
                self._environment_messaging.send_message(response, processing_time)

    def _process(self) -> Tuple[bool, EnvironmentState]:

        while self._tasks and not self._pause and not self._terminate:
            next_time = self._tasks[0][0]
            delta = next_time - self._time

            # TODO singular timestep handling
            self._time = next_time

            current_tasks: List[MessageImpl] = []
            while self._tasks and self._tasks[0][0] == self._time:
                current_tasks.append(heappop(self._tasks)[2])

            for task in current_tasks:
                if task.type == MessageType.TIMEOUT:
                    # Yay!
                    timeout = TimeoutImpl.cast_from(task.cast_to(Timeout)) #type:ignore #MYPY: Probably an issue with mypy, requires creation of helper class
                    timeout.service.process_message(task)
                else:
                    self._send_message(task)

        # Pause causes the system to stop processing and to keep task queue intact
        if self._pause:
            self._state = EnvironmentState.PAUSED

        # Terminate clears the task queue and sets the clock back to zero
        elif self._terminate:
            self._state = EnvironmentState.TERMINATED
            self._time = 0
            self._tasks.clear()

        else:
            self._state = EnvironmentState.FINISHED

        return True, self._state

    def _register_services(self) -> None:

        # First, check entry points registered via the importlib mechanism
        plugin_services = entry_points(group="cyst.services")
        for s in plugin_services:
            service_description = s.load()

            if self._service_store.get_service(service_description.name):
                print("Service with name {} already registered, skipping...".format(service_description.name))
            else:
                self._service_store.add_service(service_description)

        # Explicit addition of built-in active services
        self._service_store.add_service(firewall_service_description)

    def _register_actions(self) -> None:

        plugin_models = entry_points(group="cyst.models")
        for s in plugin_models:
            model_description = s.load()

            if model_description.namespace in self._interpreters:
                print("Behavioral model with namespace {} already registered, skipping it ...".format(
                    model_description.namespace))
            else:
                model = model_description.creation_fn(self, self._environment_resources, self._policy,
                                                      self._environment_messaging)
                self._interpreters[model_description.namespace] = model

    def _register_metadata_providers(self) -> None:

        plugin_providers = entry_points(group="cyst.metadata_providers")
        for s in plugin_providers:
            provider_description = s.load()

            if provider_description.namespace in self._metadata_providers:
                print("Metadata provider with namespace {} already registered, skipping ...".format(
                    provider_description.namespace))
            else:
                provider = provider_description.creation_fn(self._environment_resources.action_store, self)
                self._metadata_providers[provider_description.namespace] = provider
                provider.register_action_parameters()


def create_environment() -> Environment:
    e = _Environment()
    return e



from typing import Any, Type, Dict, Union, List

from cyst.core.environment.data_store_backend import DataStoreBackend, append_only_data, to_dict


class DataStoreMemoryBackend(DataStoreBackend):

    def __init__(self) -> None:
        self._store: Dict[str,Dict[str, Union[List[Dict[str, Any]], Dict[str, Any]]]] = {}

    def set(self, run_id: str, item: Any, item_type: Type) -> None:
        if run_id not in self._store:
            self._store[run_id] = {}

        item_name = item_type.__name__

        if item_type in append_only_data:
            if item_name not in self._store[run_id]:
                self._store[run_id][item_name] = []
            self._store[run_id][item_name].append(to_dict(item)) #type: ignore #append on list may in theory happen, but should not
        else:
            if item_name in self._store[run_id]:
                raise RuntimeError("Object with name {} already in the data store. Use update if you want to overwrite it".format(item_name))
            self._store[run_id][item_name] = to_dict(item)

    def get(self, run_id: str, item: Any, item_type: Type) -> Any:
        if run_id not in self._store:
            return None

        item_name = type(item).__name__
        if item_name not in self._store[run_id]:
            return None
        else:
            return self._store[run_id][item_name]

    def update(self, run_id: str, item: Any, item_type: Type) -> None:
        if run_id not in self._store:
            raise RuntimeError("Trying to update item in non-existent simulation run")
        
        if item_type in append_only_data:
            raise RuntimeError("Cannot update append-only item.")

        item_name = item_type.__name__
        if item_name not in self._store[run_id]:
            raise RuntimeError("Can't update object with name {} - not in the data store.".format(item_name))

        self._store[run_id][item_name] = to_dict(item)

    def remove(self, run_id: str, item: Any, item_type: Type) -> None:
        if run_id not in self._store:
            raise RuntimeError("Trying to remove item from non-existent simulation run")

        item_name = type(item).__name__
        if item_name not in self._store[run_id]:
            return None

        del self._store[run_id][item_name]

    def clear(self, run_id: str) -> None:
        if run_id not in self._store:
            raise RuntimeError("Cannot clear non-existent run with ID: {}".format(run_id))

        del self._store[run_id]

    def commit(self, run_id: str) -> None:
        # For in-memory store, commit is NoOp
        return



from __future__ import annotations

from heapq import heappush
from netaddr import IPAddress
from typing import TYPE_CHECKING, Optional, Any, Union

from cyst.api.environment.message import Request, Response, Status, Message
from cyst.api.environment.messaging import EnvironmentMessaging
from cyst.api.logic.access import Authorization, AuthenticationTarget, AuthenticationToken
from cyst.api.logic.action import Action
from cyst.api.logic.metadata import Metadata
from cyst.api.network.session import Session
from cyst.api.utils.counter import Counter

from cyst.core.environment.message import RequestImpl, ResponseImpl, MessageImpl, MessageType
from cyst.core.network.elements import Endpoint, InterfaceImpl
from cyst.core.network.session import SessionImpl

if TYPE_CHECKING:
    from cyst.core.environment.environment import _Environment


class EnvironmentMessagingImpl(EnvironmentMessaging):
    def __init__(self, env: _Environment):
        self._env = env

    def send_message(self, message: Message, delay: int = 0) -> None:
        m = MessageImpl.cast_from(message)
        _send_message(self._env, m, delay)

    def create_request(self, dst_ip: Union[str, IPAddress], dst_service: str = "", action: Optional[Action] = None,
                       session: Optional[Session] = None,
                       auth: Optional[Union[Authorization, AuthenticationToken]] = None) -> Request:
        return _create_request(dst_ip, dst_service, action, session, auth)

    def create_response(self, request: Request, status: Status, content: Optional[Any] = None,
                        session: Optional[Session] = None,
                        auth: Optional[Union[Authorization, AuthenticationTarget]] = None):
        return _create_response(request, status, content, session, auth)

    def open_session(self, request: Request) -> Session:
        return _open_session(self._env, request)


# ----------------------------------------------------------------------------------------------------------------------
# Free function implementations of the above class. It is being done this way to shut up the type checking and to
# overcome python's limitation on having a class implemented in multiple files.
def _create_request(dst_ip: Union[str, IPAddress], dst_service: str = "",
                    action: Action = None, session: Session = None,
                    auth: Optional[Union[Authorization, AuthenticationToken]] = None) -> Request:
    request = RequestImpl(dst_ip, dst_service, action, session, auth)
    return request


def _create_response(request: Request, status: Status, content: Optional[Any] = None,
                     session: Optional[Session] = None,
                     auth: Optional[Union[Authorization, AuthenticationTarget]] = None) -> Response:
    # Let's abuse the duck typing and "cast" Request to RequestImpl
    if isinstance(request, RequestImpl):
        response = ResponseImpl(request, status, content, session, auth)
        return response
    else:
        raise ValueError("Malformed request passed to create a response from")


def _open_session(self: _Environment, request: Request) -> Session:
    return self._network_configuration.create_session_from_message(request)


def _send_message(self, message: MessageImpl, delay: int = 0) -> None:
    # set a first hop for a message
    source = self._network.get_node_by_id(message.origin.id)
    # Find a next hop for messages without one
    if source and not message.next_hop:
        # New request with session should follow the session first
        # Response should either follow newly established session, or route to session endpoint
        # TODO rearrange it to reflect changes in response set_next_hop handling
        if message.type == MessageType.REQUEST and message.session:
            message.set_next_hop()
            # Not a pretty thing, but I am not sure how to make it better
            # it = SessionImpl.cast_from(message.session).forward_iterator
            # hop = next(it)
            # port = hop.src.port
            # iface = source.interfaces[port]

            # If this works it is a proof that the entire routing must be reviewed
            message.set_src_ip(message.path[0].src.ip)
        elif message.type == MessageType.RESPONSE:
            if message.session and message.current == SessionImpl.cast_from(message.session).endpoint:
                # This is stupid, but it complains...
                if isinstance(message, ResponseImpl):
                    message.set_in_session(True)
            message.set_next_hop()
        # Others go to a gateway
        else:
            target = message.dst_ip
            localhost = IPAddress("127.0.0.1")

            # Shortcut for localhost request
            if target == localhost:
                message.set_src_ip(localhost)
                message.set_next_hop(Endpoint(message.origin.id, 0, localhost), Endpoint(message.origin.id, 0, localhost))

            else:
                gateway, port = source.gateway(target)
                if not gateway:
                    raise Exception("Could not send a message, no gateway to route it through.")

                iface = InterfaceImpl.cast_from(source.interfaces[port])
                message.set_src_ip(iface.ip)

                message.set_origin(Endpoint(source.id, port, iface.ip))

                # First sending is specific, because the current value is set to origin
                message.set_next_hop(message.origin, iface.endpoint)

    # metadata are appended only for requests ATM. This is to test waters, as there are many different design
    # holes and things which need clarification
    if isinstance(message, Request):
        action = message.action

        if self._metadata_providers and action.namespace in self._metadata_providers:
            provider = self._metadata_providers[action.namespace]
            if provider:
                metadata = provider.get_metadata(action)
            else:
                metadata = Metadata()

            metadata.dst_ip = message.dst_ip
            metadata.dst_service = message.dst_service
            metadata.src_ip = message.src_ip
        else:
            metadata = Metadata()

        message.set_metadata(metadata)

    try:
        heappush(self._tasks, (self._time + delay, Counter().get("msg"), message))
    except Exception as e:
        self._message_log.error(f"Error sending a message, reason: {e}")

    message.sent = True

    self._message_log.debug(f"Sending a message: {str(message)}")

    if message.type is MessageType.REQUEST and f"{message.origin.id}.{message.src_service}" in self._pause_on_request:
        self._pause = True



from cyst.api.configuration.host.service import ActiveServiceConfig, PassiveServiceConfig, ServiceParameter
from cyst.api.configuration.logic.access import AccessLevel, AuthorizationConfig, AccessSchemeConfig, AuthenticationProviderConfig, AuthorizationDomainConfig, AuthorizationDomainType, FederatedAuthorizationConfig
from cyst.api.configuration.logic.data import DataConfig
from cyst.api.configuration.logic.exploit import ExploitLocality, ExploitCategory, ExploitParameterType, \
     VulnerableServiceConfig, ExploitParameterConfig, ExploitConfig
from cyst.api.configuration.network.network import NetworkConfig
from cyst.api.configuration.network.node import NodeConfig
from cyst.api.configuration.network.elements import IPAddress, IPNetwork, ConnectionConfig, PortConfig, InterfaceConfig, RouteConfig
from cyst.api.configuration.network.router import RouterConfig
from cyst.api.configuration.network.firewall import FirewallRule, FirewallChainType, FirewallChainConfig, FirewallPolicy, FirewallConfig

__all__ = ["ActiveServiceConfig", "PassiveServiceConfig", "AccessLevel", "AuthorizationConfig", "DataConfig", "ExploitLocality",
           "ExploitCategory", "ExploitParameterType", "VulnerableServiceConfig", "ExploitParameterConfig", "ExploitConfig",
           "NetworkConfig", "NodeConfig", "IPAddress", "IPNetwork", "ConnectionConfig", "PortConfig", "InterfaceConfig", "RouterConfig",
           "FirewallRule", "FirewallChainType", "FirewallChainConfig", "FirewallPolicy", "FirewallConfig", "ServiceParameter",
           "AccessSchemeConfig", "AuthenticationProviderConfig", "AuthorizationDomainConfig", "AuthorizationDomainType",
           "FederatedAuthorizationConfig", "RouteConfig"]



from abc import ABC, abstractmethod
from dataclasses import dataclass
from typing import Callable

from cyst.api.environment.configuration import ActionConfiguration
from cyst.api.environment.stores import ActionStore
from cyst.api.logic.action import Action
from cyst.api.logic.metadata import Metadata


class MetadataProvider(ABC):
    """
    Metadata providers supply messages with additional information that are meant to mimic observations that can be
    done in the real life. Two typical examples of metadata can be 1) flow information, 2) results of traffic analyses.
    """
    @abstractmethod
    def register_action_parameters(self) -> None:
        """
        There can be classes of metadata that are different for a particular action with particular parameters as
        defined by the behavioral model. These different classes can manifest depending on parameters that are not
        considered in the model itself. An example of such difference are various scanning techniques for a
        reconnaissance action (e.g., SYN, FIN, XMAS scans). Each of these techniques would have different statistical
        properties that would reflect in, e.g., flow information.

        This function gives the metadata providers an option to inject its own parameters into the actions of the action
        store. These parameters should have a closed parameter domain for this to be reasonably usable.
        Considering that this inject is done on a per-action basis, the metadata provider must understand the
        semantic of specific actions, so it is expected that such providers are released together with behavioral
        model implementation.
        """

    @abstractmethod
    def get_metadata(self, action: Action) -> Metadata:
        """
        This function should return metadata that correspond to the given action and its parameters.

        :param action: An instance of the action.
        :type action: Action

        :return: A metadata associated with the action.
        """


@dataclass
class MetadataProviderDescription:
    """ A description of a metadata provider which should be registered into the system.

    :param namespace: A namespace of actions that this provider works with.
    :type namespace: str

    :param description: A short description of the provider.
    :type description: str

    :param creation_fn: A factory function that creates instances of the metadata provider.
    :type creation_fn: Callable[[ActionStore, ActionConfiguration], MetadataProvider]
    """
    namespace: str
    description: str
    creation_fn: Callable[[ActionStore, ActionConfiguration], MetadataProvider]



from dataclasses import dataclass, field

from netaddr import IPAddress, IPNetwork
from typing import Optional
from uuid import uuid4
from serde import serialize
from serde.compat import typename

from cyst.api.configuration.configuration import ConfigItem


@serialize
@dataclass
class PortConfig(ConfigItem):
    """ Configuration of a network port.

    A network port represents an abstraction of an ethernet port, with a given IP address and a given network. A network
    port does not support a default routing through a gateway and so it is used mostly for routers, which maintain
    their own routing tables based on the port indexes.

    :param ip: The assigned IP address of the port.
    :type ip: IPAddress

    :param net: The assigned network of the port. If used only for inter-router communication, ip/32 or ip/128
        can be used.
    :type net: IPNetwork

    :param index: The index of the port. The index is used for unique addressing of a port within a node, especially
        for correctly setting routing tables. If left at the default value, it is assigned the next free index.
    :type index: int

    :param id: A unique identifier of the port configuration.
    :type id: str
    """
    ip: IPAddress = field(metadata={
        'serde_serializer': lambda x: {"cls_type": typename(type(x)), "value": str(x)},
    })
    net: IPNetwork = field(metadata={
        'serde_serializer': lambda x: {"cls_type": typename(type(x)), "value": str(x)},
    })
    index: int = field(default=-1)
    id: str = field(default_factory=lambda: str(uuid4()))


@serialize
@dataclass
class InterfaceConfig(ConfigItem):
    """ Configuration of a network interface.

    A network interface represents an abstraction of an ethernet port, with a given IP address and a given network.
    A network interface automatically calculates the gateway and therefore enables a seamless networking.

    :param ip: The assigned IP address of the interface.
    :type ip: IPAddress

    :param net: The assigned network of the interface.
    :type net: IPNetwork

    :param index: The index of the interface. The index is used for unique addressing of an interface within a node. If
        left at the default value, it is assigned the next free index.
    :type index: int

    :param id: A unique identifier of the interface configuration.
    :type id: str
    """
    ip: IPAddress = field(metadata={
        'serde_serializer': lambda x: {"cls_type": typename(type(x)), "value": str(x)}
    })
    net: IPNetwork = field(metadata={
        'serde_serializer': lambda x: {"cls_type": typename(type(x)), "value": str(x)}
    })
    index: int = field(default=-1)
    id: str = field(default_factory=lambda: str(uuid4()))


@serialize
@dataclass
class ConnectionConfig(ConfigItem):
    """ Configuration of a network connection.

    Represents a connection between two network ports/interfaces. A connection will in future support setting of
    connection properties, such as delay or packet drops. While the supporting infrastructure is partially present
    in the code now, it is not propagated into the configuration a so, each connection has a unit speed (in terms of
    the simulation time) with zero drops.

    :param src_id: The id of a source node.
    :type src_id: str

    :param src_port: The index of a source port/interface.
    :type src_port: int

    :param dst_id: The id of a destination node.
    :type dst_id: str

    :param dst_port: The index of a destination port/interface.
    :type dst_port: int

    :param id: A unique identifier of the connection configuration.
    :type id: str
    """
    src_id: str
    src_port: int
    dst_id: str
    dst_port: int
    id: str = field(default_factory=lambda: str(uuid4()))


@serialize
@dataclass
class RouteConfig(ConfigItem):
    """ Configuration of a network route.

    A route specifies which port should the traffic to specific network be routed through. Many routes can be specified
    for a node. If there is an overlap in network specification, than the resulting port is selected according to the
    route metrics (i.e., the lower the metric the higher the chance to be selected as the route). In case of a metric
    equality, the most specific network is selected.

    :param network: A network this route is related to.
    :type network: IPNetwork

    :param port: A port/interface index, where to route traffic to the particular network.
    :type port: int

    :param metric: A route metric used for deciding which route to use in case of network overlap.
    :type metric: int

    :param id: A unique identifier of the route configuration.
    :type id: str
    """
    network: IPNetwork
    port: int
    metric: int = field(default=100)
    id: str = field(default_factory=lambda: str(uuid4()))



import networkx as nx

from abc import ABC, abstractmethod
from netaddr import IPAddress
from typing import Optional, List, Dict

from cyst.api.environment.configuration import GeneralConfiguration
from cyst.core.network.elements import Connection, ConnectionImpl, Hop, Endpoint, Resolver, InterfaceImpl
from cyst.core.network.router import Router
from cyst.core.network.node import NodeImpl


# TODO: The network is largely useless after moving the object store into configuration. I should probably think about
#       whether to keep it or not
class Network(Resolver):
    def __init__(self, conf: GeneralConfiguration):
        self._graph = nx.Graph()
        self._conf = conf

    def add_node(self, node: NodeImpl) -> None:
        # Ignore already present nodes
        if self._graph.has_node(node.id):
            return

        self._graph.add_node(node.id)

    def add_connection(self, n1: NodeImpl, n1_port_index: int, n2: NodeImpl, n2_port_index: int, net: str, connection: Optional[Connection] = None) -> Connection:
        if not n1 or not n2:
            raise Exception("Could not add connection between nonexistent nodes")

        if not connection:
            connection = ConnectionImpl()

        result = True
        error = ""
        if isinstance(n1, Router):
            if isinstance(n2, Router):
                result, error = n1._connect_router(n2, connection, n2_port_index, n1_port_index)
            else:
                result, error = n1._connect_node(n2, connection, n1_port_index, n2_port_index, net)
        elif isinstance(n2, Router):
            result, error = n2._connect_node(n1, connection, n2_port_index, n1_port_index, net)
        # Direct connection
        else:
            InterfaceImpl.cast_from(n1.interfaces[n1_port_index]).connect_endpoint(
                Endpoint(n2.id, n2_port_index, n2.interfaces[n2_port_index].ip), connection)
            InterfaceImpl.cast_from(n2.interfaces[n2_port_index]).connect_endpoint(
                Endpoint(n1.id, n1_port_index, n1.interfaces[n1_port_index].ip), connection)

        if not result:
            raise Exception("Could not add connection between nodes {} and {}. Reason: {}".format(n1.id, n2.id, error))

        connection.hop = Hop(Endpoint(n1.id, n1_port_index), Endpoint(n2.id, n2_port_index))
        self._graph.add_edge(n1.id, n2.id)

        return connection

    def get_node_by_id(self, id: str = "") -> Optional[NodeImpl]:
        if not id:
            return None
        else:
            return self._conf.get_object_by_id(id, NodeImpl)

    def reset(self) -> None:
        # self._nodes_by_id.clear()
        self._graph.clear()

    def resolve_ip(self, id: str, port: int) -> IPAddress:
        node = self.get_node_by_id(id)
        if not node:
            raise ValueError("Nonexistent node id provided for resolving")

        if port >= len(node.interfaces):
            raise ValueError("Nonexistent port id provided for resolving")

        return node.interfaces[port].ip



import logging.config
import uuid
import collections

from dataclasses import dataclass, field
from typing import List, Union, Optional, Dict, Any, Type, Tuple, Callable

import jsonpickle

from cyst.api.environment.configuration import GeneralConfiguration, ObjectType, ConfigurationObjectType
from cyst.api.configuration.configuration import ConfigItem
from cyst.api.configuration.host.service import ActiveServiceConfig, PassiveServiceConfig
from cyst.api.configuration.infrastructure.infrastructure import InfrastructureConfig
from cyst.api.configuration.infrastructure.log import LogConfig, LogSource, log_defaults
from cyst.api.configuration.logic.access import AuthorizationConfig, AuthenticationProviderConfig, AccessSchemeConfig, \
    AuthorizationDomainConfig, FederatedAuthorizationConfig
from cyst.api.configuration.logic.data import DataConfig
from cyst.api.configuration.logic.exploit import VulnerableServiceConfig, ExploitParameterConfig, ExploitConfig
from cyst.api.configuration.network.elements import PortConfig, InterfaceConfig, ConnectionConfig, RouteConfig
from cyst.api.configuration.network.firewall import FirewallChainConfig, FirewallConfig, FirewallPolicy, FirewallChainType
from cyst.api.configuration.network.network import NetworkConfig
from cyst.api.configuration.network.router import RouterConfig
from cyst.api.configuration.network.node import NodeConfig
from cyst.api.environment.environment import Environment
from cyst.api.host.service import AccessLevel
from cyst.api.network.node import Node

from cyst.core.network.firewall import FirewallImpl


# ----------------------------------------------------------------------------------------------------------------------
# Converting configuration description to actual objects
# ----------------------------------------------------------------------------------------------------------------------
class Configurator:
    def __init__(self):
        self._env: Environment
        self._refs: Dict[str, Any] = {}
        self._obj_refs: Dict[str, Any] = {}
        self._connections: List[ConnectionConfig] = []
        self._nodes: List[NodeConfig] = []
        self._routers: List[RouterConfig] = []
        self._active_services: List[ActiveServiceConfig] = []
        self._passive_services: List[PassiveServiceConfig] = []
        self._firewalls: List[FirewallConfig] = []
        self._interfaces: List[Union[InterfaceConfig, PortConfig]] = []
        self._authorizations: List[AuthorizationConfig] = []
        self._data: List[DataConfig] = []
        self._exploits: List[ExploitConfig] = []
        self._authentication_providers: List[AuthenticationProviderConfig] = []
        self._access_schemes: List[AccessSchemeConfig] = []
        self._authorization_domains: List[AuthorizationDomainConfig] = []
        self._logs: List[LogConfig] = []

    def __getstate__(self):
        result = self.__dict__
        return result

    def __setstate__(self, state):
        self.__dict__.update(state)

    def reset(self):
        self._refs.clear()
        self._obj_refs.clear()
        self._connections.clear()
        self._nodes.clear()
        self._routers.clear()
        self._active_services.clear()
        self._passive_services.clear()
        self._firewalls.clear()
        self._interfaces.clear()
        self._authorizations.clear()
        self._data.clear()
        self._exploits.clear()
        self._authentication_providers.clear()
        self._access_schemes.clear()
        self._authorization_domains.clear()
        self._logs.clear()

    # ------------------------------------------------------------------------------------------------------------------
    # All these _process_XXX functions resolve nested members to their id. In the end of the preprocessing, there should
    # be only a flat configuration with ids and no nesting (see members above)
    # ------------------------------------------------------------------------------------------------------------------
    def _process_NetworkConfig(self, cfg: NetworkConfig) -> NetworkConfig:
        node_ids = []
        conn_ids = []

        for node in cfg.nodes:
            if isinstance(node, str):
                node_ids.append(node)
            else:
                node_ids.append(self._process_cfg_item(node))

        for conn in cfg.connections:
            if isinstance(conn, str):
                conn_ids.append(conn)
            else:
                conn_ids.append(self._process_cfg_item(conn))

        cfg.nodes = node_ids
        cfg.connections = conn_ids

        return cfg

    def _process_ConnectionConfig(self, cfg: ConnectionConfig) -> str:
        self._connections.append(cfg)
        self._refs[cfg.id] = cfg
        return cfg.id

    def _process_RouterConfig(self, cfg: RouterConfig):
        interface_ids = []
        traffic_processor_ids = []
        route_ids = []

        for interface in cfg.interfaces:
            if isinstance(interface, str):
                interface_ids.append(interface)
            else:
                interface_ids.append(self._process_cfg_item(interface))

        for processor in cfg.traffic_processors:
            if isinstance(processor, str):
                traffic_processor_ids.append(processor)
            else:
                traffic_processor_ids.append(self._process_cfg_item(processor))

        for route in cfg.routing_table:
            self._process_RouteConfig(route)

        cfg.interfaces = interface_ids
        cfg.traffic_processors = traffic_processor_ids

        self._routers.append(cfg)
        self._refs[cfg.id] = cfg
        return cfg.id

    def _process_NodeConfig(self, cfg: NodeConfig) -> str:
        passive_service_ids = []
        active_service_ids = []
        traffic_processor_ids = []
        interface_ids = []

        for service in cfg.passive_services:
            if isinstance(service, str):
                passive_service_ids.append(service)
            else:
                passive_service_ids.append(self._process_cfg_item(service))

        for service in cfg.active_services:
            if isinstance(service, str):
                active_service_ids.append(service)
            else:
                active_service_ids.append(self._process_cfg_item(service))
        
        for processor in cfg.traffic_processors:
            if isinstance(processor, str):
                traffic_processor_ids.append(processor)
            else:
                traffic_processor_ids.append(self._process_cfg_item(processor))

        for interface in cfg.interfaces:
            if isinstance(interface, str):
                interface_ids.append(interface)
            else:
                interface_ids.append(self._process_cfg_item(interface))

        cfg.passive_services = passive_service_ids
        cfg.active_services = active_service_ids
        cfg.traffic_processors = traffic_processor_ids
        cfg.interfaces = interface_ids

        self._nodes.append(cfg)
        self._refs[cfg.id] = cfg

        return cfg.id

    def _process_PortConfig(self, cfg: PortConfig) -> str:
        self._refs[cfg.id] = cfg
        self._interfaces.append(cfg)
        return cfg.id

    def _process_InterfaceConfig(self, cfg: InterfaceConfig) -> str:
        self._refs[cfg.id] = cfg
        self._interfaces.append(cfg)
        return cfg.id

    def _process_ActiveServiceConfig(self, cfg: ActiveServiceConfig) -> str:
        self._refs[cfg.id] = cfg
        self._active_services.append(cfg)
        return cfg.id
    
    def _process_FirewallConfig(self, cfg: FirewallConfig) -> str:
        self._refs[cfg.id] = cfg
        self._firewalls.append(cfg)
        return cfg.id

    def _process_PassiveServiceConfig(self, cfg: PassiveServiceConfig) -> str:
        self._refs[cfg.id] = cfg
        self._passive_services.append(cfg)

        public_data_ids = []
        private_data_ids = []
        public_auth_ids = []
        private_auth_ids = []

        for data in cfg.public_data:
            if isinstance(data, str):
                public_data_ids.append(data)
            else:
                public_data_ids.append(self._process_cfg_item(data))

        for data in cfg.private_data:
            if isinstance(data, str):
                private_data_ids.append(data)
            else:
                private_data_ids.append(self._process_cfg_item(data))

        for auth in cfg.public_authorizations:
            if isinstance(auth, str):
                public_auth_ids.append(auth)
            else:
                public_auth_ids.append(self._process_cfg_item(auth))

        for auth in cfg.private_authorizations:
            if isinstance(auth, str):
                private_auth_ids.append(auth)
            else:
                private_auth_ids.append(self._process_cfg_item(auth))

        cfg.public_data = public_data_ids
        cfg.private_data = private_data_ids
        cfg.public_authorizations = public_auth_ids
        cfg.private_authorizations = private_auth_ids

        auth_provider_ids = []
        for provider in cfg.authentication_providers:
            if isinstance(provider, str):
                auth_provider_ids.append(provider)
            else:
                auth_provider_ids.append(self._process_cfg_item(provider))
        cfg.authentication_providers = auth_provider_ids

        access_scheme_ids = []
        for scheme in cfg.access_schemes:
            if isinstance(scheme, str):
                access_scheme_ids.append(scheme)
            else:
                access_scheme_ids.append(self._process_cfg_item(scheme))
        cfg.access_schemes = access_scheme_ids

        return cfg.id

    def _process_AuthorizationConfig(self, cfg: AuthorizationConfig) -> str:
        self._refs[cfg.id] = cfg
        self._authorizations.append(cfg)
        return cfg.id

    def _process_DataConfig(self, cfg: DataConfig) -> str:
        self._refs[cfg.id] = cfg
        self._data.append(cfg)
        return cfg.id

    def _process_ExploitConfig(self, cfg: ExploitConfig) -> str:
        self._refs[cfg.id] = cfg
        self._exploits.append(cfg)
        return cfg.id

    def _process_AuthenticationProviderConfig(self, cfg: AuthenticationProviderConfig) -> str:
        self._refs[cfg.id] = cfg
        self._authentication_providers.append(cfg)
        return cfg.id

    def _process_AccessSchemeConfig(self, cfg: AccessSchemeConfig) -> str:

        auth_provider_ids = []
        for provider in cfg.authentication_providers:
            if isinstance(provider, str):
                auth_provider_ids.append(provider)
            else:
                auth_provider_ids.append(self._process_cfg_item(provider))
        cfg.authentication_providers = auth_provider_ids

        if not isinstance(cfg.authorization_domain, str):
            cfg.authorization_domain = self._process_cfg_item(cfg.authorization_domain)

        self._refs[cfg.id] = cfg
        self._access_schemes.append(cfg)
        return cfg.id

    def _process_AuthorizationDomainConfig(self, cfg: AuthorizationDomainConfig) -> str:

        authorization_ids = []
        for auth in cfg.authorizations:
            if isinstance(auth, str):
                authorization_ids.append(auth)
            else:
                authorization_ids.append(self._process_cfg_item(auth))
        cfg.authorizations = authorization_ids

        self._refs[cfg.id] = cfg
        self._authorization_domains.append(cfg)
        return cfg.id

    def _process_FederatedAuthorizationConfig(self, cfg: FederatedAuthorizationConfig):

        # TODO : Ask how this is meant to be handled

        self._refs[cfg.id] = cfg
        return cfg.id

    def _process_RouteConfig(self, cfg: RouteConfig) -> str:

        self._refs[cfg.id] = cfg
        return cfg.id

    def _process_LogConfig(self, cfg: LogConfig) -> str:
        self._refs[cfg.id] = cfg
        self._logs.append(cfg)
        return cfg.id

    def _process_InfrastructureConfig(self, cfg: InfrastructureConfig) -> str:
        self._refs[cfg.id] = cfg
        for log in cfg.log:
            self._process_LogConfig(log)
        return cfg.id

    def _process_default(self, cfg):
        raise ValueError("Unknown config type provided")

    def _process_cfg_item(self, cfg: Any) -> str:
        if hasattr(cfg, "id") and cfg.id in self._refs:
            raise ValueError("Duplicate identifier found: {}".format(cfg.id))

        fn: Callable[[ConfigItem], str] = getattr(self, "_process_" + type(cfg).__name__, self._process_default)
        return fn(cfg)

    def get_configuration_by_id(self, id: str) -> Optional[Any]:
        if id not in self._refs:
            return None
        else:
            return self._refs[id]

    def get_object_by_id(self, id: str) -> Optional[Any]:
        if id not in self._obj_refs:
            return None
        else:
            return self._obj_refs[id]

    def get_objects_by_type(self, type: Type[ObjectType]) -> List[ObjectType]:
        result = []
        for obj in self._obj_refs.values():
            if isinstance(obj, type):
                result.append(obj)
        return result

    def _resolve_config_item(self, item: ConfigItem) -> ConfigItem:
        replaced = {}
        for key, value in item.__dict__.items():
            if isinstance(value, str) and not key.endswith("id") and value in self._refs:
                replaced[key] = self._resolve_config_item(self._refs[value])
            if isinstance(value, list):
                tmp = []
                for element in value:
                    if isinstance(element, str) and element in self._refs:
                        tmp.append(self._resolve_config_item(self._refs[element]))
                if tmp:
                    replaced[key] = tmp
        item.__dict__.update(replaced)
        return item

    def get_configuration(self) -> List[ConfigItem]:
        top_level = [*self._nodes, *self._routers, *self._connections, *self._exploits]
        result = []
        for item in top_level:
            result.append(self._resolve_config_item(item))
        return result

    def add_object(self, id: str, obj: Any) -> None:
        if id in self._obj_refs:
            tmp = self._obj_refs[id]
            raise RuntimeError(f"Attempting to add an already existing object with id: {id} and type: {type(tmp)}")

        self._obj_refs[id] = obj

    # ------------------------------------------------------------------------------------------------------------------
    # Gather all configuration items
    # TODO: Return of Environment is an artifact of previous development. Need to change it to something more useful
    def configure(self, env: Optional[Environment],
                  *configs: Union[NetworkConfig, ConnectionConfig, RouterConfig, NodeConfig,
                                  InterfaceConfig, ActiveServiceConfig, PassiveServiceConfig,
                                  AuthorizationConfig, DataConfig]) -> Environment:
        if not env:
            self._env = Environment.create()
        else:
            self._env = env
            # Everytime a configuration of environment is required, there must be a complete reset of it
            env.control.reset()

        # --------------------------------------------------------------------------------------------------------------
        # Process all provided items and do a complete id->cfg mapping
        for cfg in configs:
            self._process_cfg_item(cfg)

        # --------------------------------------------------------------------------------------------------------------
        # Begin by processing the infrastructure configuration
        # Logs:
        # Get defaults
        log_configs = {}
        for log in log_defaults:
            log_configs[log.source] = log

        # Overwrite with user configuration
        for cfg in self._logs:
            log_configs[cfg.source] = cfg

        formatters = {
            '__default': {
                'format': "[%(asctime)s] :: %(name)s — %(levelname)s :: %(message)s"
            }
        }

        handlers = {}
        loggers = {}

        log_source_map = {
            LogSource.SYSTEM: 'system',
            LogSource.MESSAGING: 'messaging',
            LogSource.MODEL: 'model.',
            LogSource.SERVICE: 'service.'
        }

        for cfg in log_configs.values():
            if not cfg.log_console and not cfg.log_file:
                continue

            handler_console = {}
            handler_file = {}

            if cfg.log_console:
                handler_console = {
                    'class': 'logging.StreamHandler',
                    'formatter': '__default',
                    'level': cfg.log_level,
                    'stream': 'ext://sys.stdout'
                }

            if cfg.log_file and cfg.file_path:
                handler_file = {
                    'class': 'logging.FileHandler',
                    'formatter': '__default',
                    'level': cfg.log_level,
                    'filename': cfg.file_path
                }

            if handler_console or handler_file:
                handler_list = []
                if handler_console:
                    id = "__console_" + log_source_map[cfg.source]
                    handlers[id] = handler_console
                    handler_list.append(id)
                if handler_file:
                    id = "__file_" + log_source_map[cfg.source]
                    handlers[id] = handler_file
                    handler_list.append(id)

                loggers[log_source_map[cfg.source]] = {
                    'handlers': handler_list,
                    'level': cfg.log_level
                }

        log_config_dict = {
            'version': 1,
            'formatters': formatters,
            'handlers': handlers,
            'loggers': loggers
        }

        logging.config.dictConfig(log_config_dict)

        # --------------------------------------------------------------------------------------------------------------
        # Now that each configuration item is accounted for, build it.
        # Build order:
        # 1) Authorizations, Data and Exploits
        # 2) Passive Services
        # 3) Interfaces
        # 4) Nodes and routers
        # 5) Connections

        # 1) Authorizations, Data and Exploits
        # for auth in self._authorizations:
        #    a = self._env.policy.create_authorization(auth.identity, auth.nodes, auth.services, auth.access_level, auth.id)
        #    self._env.policy.add_authorization(a)
        #    self._obj_refs[auth.id] = a

        # Building authentication and authorization infrastructure
        # Prepare authentication tokens based on the authorizations in access schemes
        for scheme in self._access_schemes:

            scheme_instance = self._env.configuration.access.create_access_scheme(scheme.id)

            authorization_domain = self._refs[scheme.authorization_domain]

            # Go through all providers and if they do not exist instantiate them
            for provider_id in scheme.authentication_providers:
                provider_conf: AuthenticationProviderConfig = self._refs[provider_id]

                if provider_id not in self._obj_refs:
                    provider = self._env.configuration.access.create_authentication_provider(
                        provider_conf.provider_type,
                        provider_conf.token_type,
                        provider_conf.token_security,
                        provider_conf.ip,
                        provider_conf.timeout,
                        provider_id
                    )
                else:
                    provider = self._obj_refs[provider_id]

                self._env.configuration.access.add_provider_to_scheme(provider, scheme_instance)

            for auth_id in authorization_domain.authorizations:
                auth = self._refs[auth_id]
                if isinstance(auth, AuthorizationConfig) or isinstance(auth, FederatedAuthorizationConfig):
                    identity = auth.identity

                    # This is not very pretty, but the original version that ran this code for each provider caused
                    # problems with the new approach that correctly tracks IDs of objects. In this case objects with
                    # identical IDs could be created multiple times.
                    for provider_id in scheme.authentication_providers:
                        self._env.configuration.access.create_and_register_authentication_token(self._obj_refs[provider_id], identity)

                    # WARN: this creates authorizations with services = ["*"], nodes = ["*"] if not Federated is
                    # created, when the authorization process is
                    # done, these are only used as templates so should not be a problem
                    authorization = self._env.configuration.access.create_authorization(auth.identity,
                                                                                        auth.access_level, auth.id)\
                        if not isinstance(auth, FederatedAuthorizationConfig) else \
                        self._env.configuration.access.create_authorization(auth.identity, auth.access_level,
                                                                            auth.id, auth.nodes, auth.services)

                    self._env.configuration.access.add_authorization_to_scheme(authorization, scheme_instance)
                else:
                    raise RuntimeError("Wrong object type provided instead of (Federated)AuthorizationConfig")

        for data in self._data:
            d = self._env.configuration.service.create_data(data.id, data.owner, data.description)

        for exploit in self._exploits:
            params = []
            if exploit.parameters:
                for p in exploit.parameters:
                    param = self._env.configuration.exploit.create_exploit_parameter(p.type, p.value, p.immutable)
                    params.append(param)

            services = []
            for s in exploit.services:
                service = self._env.configuration.exploit.create_vulnerable_service(s.name, s.min_version,
                                                                                    s.max_version)
                services.append(service)

            e = self._env.configuration.exploit.create_exploit(exploit.id, services, exploit.locality, exploit.category,
                                                               *params)

            self._env.configuration.exploit.add_exploit(e)

        # 2) Passive Services
        passive_service_obj = {}
        for service in self._passive_services:
            s = self._env.configuration.service.create_passive_service(service.type, service.owner, service.version,
                                                                       service.local, service.access_level, service.id)

            for d in service.public_data:
                self._env.configuration.service.public_data(s.passive_service).append(self._obj_refs[d])
            for d in service.private_data:
                self._env.configuration.service.private_data(s.passive_service).append(self._obj_refs[d])
            for a in service.public_authorizations:
                self._env.configuration.service.public_authorizations(s.passive_service).append(self._obj_refs[a])
            for a in service.private_authorizations:
                self._env.configuration.service.private_authorizations(s.passive_service).append(self._obj_refs[a])

            for p in service.parameters:
                self._env.configuration.service.set_service_parameter(s.passive_service, p[0], p[1])

            for prov in service.authentication_providers:
                self._env.configuration.service.provides_auth(s,
                                                              self._obj_refs[prov.id if isinstance(prov,
                                                                            AuthenticationProviderConfig) else prov])

            for scheme in service.access_schemes:
                self._env.configuration.service.set_scheme(s, self._obj_refs[
                    scheme.id if isinstance(scheme, AccessSchemeConfig) else scheme])

            passive_service_obj[service.id] = s

        # 3) Interfaces
        for iface in self._interfaces:
            if isinstance(iface, PortConfig):
                self._env.configuration.node.create_port(iface.ip, str(iface.net.netmask), iface.index, iface.id)
            else:
                # TODO: Missing a setting of a gateway (Really todo?)
                self._env.configuration.node.create_interface(iface.ip, str(iface.net.netmask), iface.index, iface.id)

        # 4) Nodes
        for node in self._nodes:
            n = self._env.configuration.node.create_node(node.id)
            for i in node.interfaces:
                obj_i = self._obj_refs[i]
                self._env.configuration.node.add_interface(n, obj_i, obj_i.index)

            for service in node.passive_services:
                self._env.configuration.node.add_service(n, passive_service_obj[service])

            for service in node.active_services:
                service_cfg: ActiveServiceConfig = self._refs[service]
                s = self._env.configuration.service.create_active_service(service_cfg.type, service_cfg.owner,
                                                                          service_cfg.name, n, service_cfg.access_level,
                                                                          service_cfg.configuration, service_cfg.id)
                self._env.configuration.node.add_service(n, s)
            
            for processor in node.traffic_processors:
                processor_cfg: Union[ActiveServiceConfig, FirewallConfig] = self._refs[processor]
                if isinstance(processor_cfg, ActiveServiceConfig):
                    s = self._env.configuration.service.create_active_service(processor_cfg.type, processor_cfg.owner,
                                                                              processor_cfg.name, n, processor_cfg.access_level,
                                                                              processor_cfg.configuration)
                else:
                    # TODO: Owner/name/access_level not in the config. It should probably be there. Otherwise we have
                    #       to hardcode.
                    s = self._env.configuration.service.create_active_service("firewall", "root",
                                                                              "firewall", n, AccessLevel.ELEVATED,
                                                                              None)

                    # This is not very pretty. But coding around it would require unnecessary expansion of the API.
                    if isinstance(s.active_service, FirewallImpl):
                        impl: FirewallImpl = s.active_service

                        for chain in processor_cfg.chains:
                            impl.set_default_policy(chain.type, chain.policy)
                            for rule in chain.rules:
                                impl.add_rule(chain.type, rule)

                self._env.configuration.node.add_traffic_processor(n, s.active_service)

            self._env.configuration.node.set_shell(n, n.services.get(node.shell, None))

            self._env.configuration.network.add_node(n)

        for router in self._routers:
            r = self._env.configuration.node.create_router(router.id, self._env.messaging)
            # TODO: This one complains that it is the List of Interface configs, while, after config processing it is
            #       a list of strings. This should probably be fixed to keep it type-happy.
            for iface_id in router.interfaces:
                iface = self._obj_refs[iface_id]
                self._env.configuration.node.add_interface(r, iface, iface.index)

            for route in router.routing_table:
                route_obj = self._env.configuration.node.create_route(route.network, route.port, route.metric)
                self._env.configuration.node.add_route(r, route_obj)

            # TODO: Code duplication... this is one of the things to fix once routers and nodes are combined together

            # HACK: To have a sensible default, we add a permissive firewall that enables forwarding on the router if
            # there are no traffic processors specified. The implementation is not pretty.
            have_fw = False
            for processor in router.traffic_processors:
                processor_cfg: Union[ActiveServiceConfig, FirewallConfig] = self._refs[processor]
                if isinstance(processor_cfg, FirewallConfig):
                    have_fw = True

            if not have_fw:
                fw_id = str(uuid.uuid4())
                self._refs[fw_id] = FirewallConfig(
                                        default_policy=FirewallPolicy.DENY,
                                        chains=[
                                            FirewallChainConfig(
                                                type=FirewallChainType.FORWARD,
                                                policy=FirewallPolicy.ALLOW,
                                                rules=[]
                                            )
                                        ]
                                    )
                router.traffic_processors.append(fw_id)

            for processor in router.traffic_processors:
                processor_cfg: Union[ActiveServiceConfig, FirewallConfig] = self._refs[processor]
                if isinstance(processor_cfg, ActiveServiceConfig):
                    s = self._env.configuration.service.create_active_service(processor_cfg.type, processor_cfg.owner,
                                                                              processor_cfg.name, r, processor_cfg.access_level,
                                                                              processor_cfg.configuration)
                else:
                    # TODO: Owner/name/access_level not in the config. It should probably be there. Otherwise we have
                    #       to hardcode.
                    s = self._env.configuration.service.create_active_service("firewall", "root",
                                                                              "firewall", r, AccessLevel.ELEVATED,
                                                                              {"default_policy": processor_cfg.default_policy})

                    # This is not very pretty. But coding around it would require unnecessary expansion of the API.
                    if isinstance(s.active_service, FirewallImpl):
                        impl: FirewallImpl = s.active_service

                        for chain in processor_cfg.chains:
                            impl.set_default_policy(chain.type, chain.policy)
                            for rule in chain.rules:
                                impl.add_rule(chain.type, rule)

                self._env.configuration.node.add_traffic_processor(r, s.active_service)

            self._env.configuration.network.add_node(r)

        # 5) Connections
        for conn in self._connections:
            src = self._obj_refs[conn.src_id]
            dst = self._obj_refs[conn.dst_id]

            self._env.configuration.network.add_connection(src, dst, conn.src_port, conn.dst_port)

        return self._env


# ----------------------------------------------------------------------------------------------------------------------
class GeneralConfigurationImpl(GeneralConfiguration):

    def __init__(self, env: Environment) -> None:
        self._env = env
        self._configurator = Configurator()

    def configure(self, *config_item: Union[NetworkConfig, ConnectionConfig, RouterConfig, NodeConfig,
                                            InterfaceConfig, ActiveServiceConfig, PassiveServiceConfig,
                                            AuthorizationConfig, DataConfig]) -> Environment:
        self._configurator.reset()
        self._configurator.configure(self._env, *config_item)
        return self._env

    def get_configuration(self) -> List[ConfigItem]:
        return self._configurator.get_configuration()

    def save_configuration(self, indent: Optional[int]) -> str:
        return jsonpickle.encode(self._configurator.get_configuration(), make_refs=False, indent=indent)

    def load_configuration(self, config: str) -> List[ConfigItem]:
        return jsonpickle.decode(config)

    def get_configuration_by_id(self, id: str,
                                configuration_type: Type[ConfigurationObjectType]) -> ConfigurationObjectType:

        c = self._configurator.get_configuration_by_id(id)
        if not isinstance(c, configuration_type):
            raise AttributeError(
                "Attempting to cast configuration object with id: {} to an incompatible type: {}".format(id,
                                                                                                         str(configuration_type)))
        return c

    def get_object_by_id(self, id: str, object_type: Type[ObjectType]) -> ObjectType:

        o = self._configurator.get_object_by_id(id)
        if not isinstance(o, object_type):
            raise AttributeError(
                "Attempting to cast object with id: {} to an incompatible type: {}. Type is {}".format(id,
                                                                                                       str(object_type),
                                                                                                       type(o)))
        return o

    def get_objects_by_type(self, type: Type[ObjectType]) -> List[ObjectType]:
        return self._configurator.get_objects_by_type(type)

    def add_object(self, id: str, obj: Any) -> None:
        self._configurator.add_object(id, obj)

    @staticmethod
    def cast_from(o: GeneralConfiguration) -> 'GeneralConfigurationImpl':
        if isinstance(o, GeneralConfigurationImpl):
            return o
        else:
            raise ValueError("Malformed underlying object passed with the GeneralConfiguration interface")


# ----------------------------------------------------------------------------------------------------------------------
# Runtime configuration of the environment. Can be filled from different sources
@dataclass
class RuntimeConfiguration:
    data_backend: str = "MEMORY"
    data_backend_params: Dict[str, str] = field(default_factory=lambda: {})
    run_id: str = field(default_factory=lambda: str(uuid.uuid4()))
    config_id: str = ""
    config_filename: str = ""




from abc import ABC, abstractmethod
from typing import List, Optional, Tuple, Union

from cyst.api.environment.message import Message
from cyst.api.logic.action import Action, ActionDescription
from cyst.api.logic.exploit import Exploit, ExploitCategory
from cyst.api.network.node import Node


class ActionStore(ABC):
    """
    Action store provides an access to actions that are available to services.
    """

    @abstractmethod
    def get(self, id: str = "") -> Optional[Action]:
        """
        Returns an action with given ID. This function makes a copy of the object, which is present in the store. This
        is a preferred variant, because any parameters set on that action would propagate to the store.

        :param id: A unique ID of the action.
        :type id: str

        :return: An action, if there is one with such ID.
        """

    @abstractmethod
    def get_ref(self, id: str = "") -> Optional[Action]:
        """
        Return an action with give ID. This function returns a reference to the object stored in the store and any
        parameter alterations will propagate to all subsequent queries for this action.

        :param id: A unique ID of the action.
        :type id: str

        :return: An action, if there is one with such ID.
        """

    @abstractmethod
    def get_prefixed(self, prefix: str = "") -> List[Action]:
        """
        Gets a list of actions, whose ID starts with a given string. This is usually done to get access to the entire
        namespace of a particular behavioral model.

        The list will contain copies of actions present in the store. Getting multiple references in one call is not
        supported.

        :param prefix: The prefix all actions IDs must share.
        :type prefix: str

        :return: A list of actions with the same prefix.
        """

    @abstractmethod
    def add(self, action: ActionDescription) -> None:
        """
        Adds a new action to the store. This function should be used in two cases:

        * Adding new action for a behavioral model. Such action must have a processing function mapped to the
          action ID.

        * Adding new action for intra-agent communication. There is no requirement on the action form, however
          an exception will be thrown, if this action is directed to a passive service, as the system will have
          no idea how to process it.

        :param action: A description of the action to add.
        :type action: ActionDescription

        :return: None
        """


class ExploitStore(ABC):
    """
    Exploit store provides an access to exploits that can be used together with actions. Unlike the action store,
    runtime definition of exploits by services is not permitted. This must be done through the
    :class:`cyst.api.environment.configuration.ExploitConfiguration` interface.
    """

    @abstractmethod
    def get_exploit(self, id: str = "", service: str = "", category: ExploitCategory = ExploitCategory.NONE) -> Optional[List[Exploit]]:
        """
        Gets an exploit, which satisfy all the parameters.

        :param id: An explicit ID of an exploit.
        :type id: str

        :param service: An ID of a service the exploit can be used at.
        :type service: str

        :param category: A category that the exploit should have. If the ExploitCategory.NONE is set, then the category
            is not considered when retrieving the exploits.
        :type category: ExploitCategory

        :return: A list of exploits satisfying the parameters.
        """

    @abstractmethod
    def evaluate_exploit(self, exploit: Union[str, Exploit], message: Message, node: Node) -> Tuple[bool, str]:
        """
        Evaluates, whether the provided exploit is applicable, given the message which carries the relevant action and
        a concrete node. TODO: This interface is cumbersome. While this is best fit for the data that interpreters
        receive, it is confusing at best.

        :param exploit: The ID of the exploit or its instance.
        :type exploit: Union[str, Exploit]

        :param message: An instance of the message which carried the exploit.
        :type message: Message

        :param node: An instance of the node, where the exploit is being applied.
        :type node: Node

        :return: (True, _) if exploit is applicable, (False, reason) otherwise.
        """



from typing import Tuple, Callable

from cyst.api.environment.configuration import EnvironmentConfiguration
from cyst.api.environment.interpreter import ActionInterpreter, ActionInterpreterDescription
from cyst.api.environment.message import Request, Response, Status, StatusOrigin, StatusValue
from cyst.api.environment.messaging import EnvironmentMessaging
from cyst.api.environment.policy import EnvironmentPolicy
from cyst.api.environment.resources import EnvironmentResources
from cyst.api.logic.action import ActionDescription, ActionToken, ActionParameterType, ActionParameterDomain, ActionParameterDomainType, ActionParameter
from cyst.api.network.node import Node


class CYSTModel(ActionInterpreter):
    def __init__(self, configuration: EnvironmentConfiguration, resources: EnvironmentResources,
                 policy: EnvironmentPolicy, messaging: EnvironmentMessaging) -> None:

        self._configuration = configuration
        self._action_store = resources.action_store
        self._exploit_store = resources.exploit_store
        self._policy = policy
        self._messaging = messaging

        self._action_store.add(ActionDescription("cyst:test:echo_success",
                                                 "A testing message that returns a SERVICE|SUCCESS",
                                                 [ActionParameter(ActionParameterType.NONE, "punch_strength",
                                                                  configuration.action.create_action_parameter_domain_options("weak", ["weak", "super strong"]))],
                                                 [(ActionToken.NONE, ActionToken.NONE)]))  # No tokens

        self._action_store.add(ActionDescription("cyst:test:echo_failure",
                                                 "A testing message that returns a SERVICE|FAILURE",
                                                 [],  # No parameters
                                                 [(ActionToken.NONE, ActionToken.NONE)]))  # No tokens

        self._action_store.add(ActionDescription("cyst:test:echo_error",
                                                 "A testing message that returns a SERVICE|ERROR",
                                                 [],  # No parameters
                                                 [(ActionToken.NONE, ActionToken.NONE)]))  # No tokens

        self._action_store.add(ActionDescription("cyst:network:create_session",
                                                 "Create a session to a destination service",
                                                 [],  # No parameters
                                                 [(ActionToken.NONE, ActionToken.NONE)]))  # No tokens

        self._action_store.add(ActionDescription("cyst:host:get_services",
                                                 "Get list of services on target node",
                                                 [],  # No parameters
                                                 [(ActionToken.NONE, ActionToken.NONE)]))  # No tokens

        self._action_store.add(ActionDescription("cyst:host:get_remote_services",
                                                 "Get list of services on target node",
                                                 [],  # No parameters
                                                 [(ActionToken.NONE, ActionToken.NONE)]))  # No tokens

        self._action_store.add(ActionDescription("cyst:host:get_local_services",
                                                 "Get list of services on target node",
                                                 [],  # No parameters
                                                 [(ActionToken.NONE, ActionToken.NONE)]))  # No tokens

        self._action_store.add(ActionDescription("cyst:compound:session_after_exploit",
                                                 "Create a session after a successful application of an exploit",
                                                 [],  # No parameters
                                                 [(ActionToken.NONE, ActionToken.NONE)]))  # No tokens

        self._action_store.add(ActionDescription("cyst:active_service:action_1",
                                                 "A placeholder action for active services instead of dedicated behavioral model.",
                                                 [],  # No parameters
                                                 [(ActionToken.NONE, ActionToken.NONE)]))  # No tokens

        self._action_store.add(ActionDescription("cyst:active_service:action_2",
                                                 "A placeholder action for active services instead of dedicated behavioral model.",
                                                 [],  # No parameters
                                                 [(ActionToken.NONE, ActionToken.NONE)]))  # No tokens

        self._action_store.add(ActionDescription("cyst:active_service:action_3",
                                                 "A placeholder action for active services instead of dedicated behavioral model.",
                                                 [],  # No parameters
                                                 [(ActionToken.NONE, ActionToken.NONE)]))  # No tokens

        self._action_store.add(ActionDescription("cyst:active_service:open_session",
                                                 "Open a session to an existing active service acting as forward/reverse shell.",
                                                 [],  # No parameters
                                                 [(ActionToken.NONE, ActionToken.NONE)]))  # No tokens

    def evaluate(self, message: Request, node: Node) -> Tuple[int, Response]:
        if not message.action:
            raise ValueError("Action not provided")

        action_name = "_".join(message.action.fragments)
        fn: Callable[[Request, Node], Tuple[int, Response]] = getattr(self, "process_" + action_name, self.process_default)
        return fn(message, node)

    # ------------------------------------------------------------------------------------------------------------------
    # CYST:TEST
    def process_default(self, message: Request, node: Node) -> Tuple[int, Response]:
        print("Could not evaluate message. Tag in `cyst` namespace unknown. " + str(message))
        return 0, self._messaging.create_response(message, status=Status(StatusOrigin.SYSTEM, StatusValue.ERROR), session=message.session)

    def process_test_echo_success(self, message: Request, node: Node) -> Tuple[int, Response]:
        return 1, self._messaging.create_response(message, status=Status(StatusOrigin.SERVICE, StatusValue.SUCCESS),
                                                  session=message.session, auth=message.auth)

    def process_test_echo_failure(self, message: Request, node: Node) -> Tuple[int, Response]:
        return 1, self._messaging.create_response(message, status=Status(StatusOrigin.SERVICE, StatusValue.FAILURE),
                                                  session=message.session, auth=message.auth)

    def process_test_echo_error(self, message: Request, node: Node) -> Tuple[int, Response]:
        return 1, self._messaging.create_response(message, status=Status(StatusOrigin.SERVICE, StatusValue.ERROR),
                                                  session=message.session, auth=message.auth)

    # ------------------------------------------------------------------------------------------------------------------
    # CYST:NETWORK
    def process_network_create_session(self, message: Request, node: Node) -> Tuple[int, Response]:
        session = self._configuration.network.create_session_from_message(message)
        return 1, self._messaging.create_response(message, status=Status(StatusOrigin.SERVICE, StatusValue.SUCCESS),
                                                  session=session, auth=message.auth)

    # ------------------------------------------------------------------------------------------------------------------
    # CYST:HOST
    def process_host_get_services(self, message: Request, node: Node) -> Tuple[int, Response]:
        services = []
        for service in node.services.values():
            if service.passive_service:
                services.append((service.name, service.passive_service.version))
        return 1, self._messaging.create_response(message, status=Status(StatusOrigin.SERVICE, StatusValue.SUCCESS),
                                                  session=message.session, auth=message.auth, content=services)

    def process_host_get_remote_services(self, message: Request, node: Node) -> Tuple[int, Response]:
        services = []
        for service in node.services.values():
            if service.passive_service and not service.passive_service.local:
                services.append((service.name, service.passive_service.version))
        return 1, self._messaging.create_response(message, status=Status(StatusOrigin.SERVICE, StatusValue.SUCCESS),
                                                  session=message.session, auth=message.auth, content=services)

    def process_host_get_local_services(self, message: Request, node: Node) -> Tuple[int, Response]:
        services = []
        for service in node.services.values():
            if service.passive_service and service.passive_service.local:
                services.append((service.name, service.passive_service.version))
        return 1, self._messaging.create_response(message, status=Status(StatusOrigin.SERVICE, StatusValue.SUCCESS),
                                                  session=message.session, auth=message.auth, content=services)

    # ------------------------------------------------------------------------------------------------------------------
    # CYST:COMPOUND
    def process_compound_session_after_exploit(self, message: Request, node: Node) -> Tuple[int, Response]:
        # TODO: Add check if exploit category and locality is ok
        # Check if the service is running on the target
        error = ""
        if not message.dst_service:
            error = "Service for session creation not specified"
        # and that an exploit is provided
        elif not message.action.exploit:
            error = "Exploit not specified to ensure session creation"
        # and it actually works
        elif not self._exploit_store.evaluate_exploit(message.action.exploit, message, node):
            error = f"Service {message.dst_service} not exploitable using the exploit {message.action.exploit.id}"

        if error:
            return 1, self._messaging.create_response(message, Status(StatusOrigin.NODE, StatusValue.ERROR), error,
                                                      session=message.session)
        else:
            return 1, self._messaging.create_response(message, Status(StatusOrigin.SERVICE, StatusValue.SUCCESS),
                                                      session=self._configuration.network.create_session_from_message(
                                                          message),
                                                      auth=message.auth)

    # ------------------------------------------------------------------------------------------------------------------
    # CYST:ACTIVE_SERVICE
    def process_active_service_action_1(self, message: Request, node: Node) -> Tuple[int, Response]:
        # These actions cannot be called on passive services
        return 1, self._messaging.create_response(message, Status(StatusOrigin.SYSTEM, StatusValue.ERROR),
                                                  "Cannot call active service placeholder actions on passive services.",
                                                  session=message.session)

    def process_active_service_action_2(self, message: Request, node: Node) -> Tuple[int, Response]:
        # These actions cannot be called on passive services
        return 1, self._messaging.create_response(message, Status(StatusOrigin.SYSTEM, StatusValue.ERROR),
                                                  "Cannot call active service placeholder actions on passive services.",
                                                  session=message.session)

    def process_active_service_action_3(self, message: Request, node: Node) -> Tuple[int, Response]:
        # These actions cannot be called on passive services
        return 1, self._messaging.create_response(message, Status(StatusOrigin.SYSTEM, StatusValue.ERROR),
                                                  "Cannot call active service placeholder actions on passive services.",
                                                  session=message.session)

    def process_active_service_open_session(self, message: Request, node: Node):
        # These actions cannot be called on passive services
        return 1, self._messaging.create_response(message, Status(StatusOrigin.SYSTEM, StatusValue.ERROR),
                                                  "Cannot open session with service that's not an active shell service",
                                                  session=message.session)


def create_cyst_model(configuration: EnvironmentConfiguration, resources: EnvironmentResources,
                      policy: EnvironmentPolicy, messaging: EnvironmentMessaging) -> ActionInterpreter:
    model = CYSTModel(configuration, resources, policy, messaging)
    return model


action_interpreter_description = ActionInterpreterDescription(
    "cyst",
    "Behavioral model that is equivalent to CYST actionable API",
    create_cyst_model
)



from __future__ import annotations

from heapq import heappush
from time import struct_time
from typing import TYPE_CHECKING, Any

from cyst.api.environment.clock import Clock
from cyst.api.environment.resources import EnvironmentResources
from cyst.api.environment.stats import Statistics
from cyst.api.environment.stores import ExploitStore, ActionStore
from cyst.api.host.service import ActiveService
from cyst.api.utils.counter import Counter

from cyst.core.environment.message import TimeoutImpl
from cyst.core.environment.stats import StatisticsImpl
from cyst.core.environment.stores import ActionStoreImpl, ServiceStoreImpl, ExploitStoreImpl

if TYPE_CHECKING:
    from cyst.core.environment.environment import _Environment


# ----------------------------------------------------------------------------------------------------------------------
class _Clock(Clock):
    def __init__(self, env: _Environment):
        self._env = env

    def simulation_time(self) -> int:
        return _simulation_time(self._env)

    def hybrid_time(self) -> struct_time:
        return _hybrid_time(self._env)

    def timeout(self, service: ActiveService, delay: int, content: Any) -> None:
        return _timeout(self._env, service, delay, content)


def _simulation_time(self: _Environment) -> int:
    return self._time


def _hybrid_time(self: _Environment) -> struct_time:
    # TODO this should be local time + self._time miliseconds
    return self._start_time


def _timeout(self: _Environment, service: ActiveService, delay: int, content: Any) -> None:
    m = TimeoutImpl(service, self._time, delay, content)
    heappush(self._tasks, (self._time + delay, Counter().get("msg"), m))


# ----------------------------------------------------------------------------------------------------------------------
class EnvironmentResourcesImpl(EnvironmentResources):
    def __init__(self, env: _Environment):
        self._env = env
        self._action_store = ActionStoreImpl()
        self._exploit_store = ExploitStoreImpl()
        self._clock = _Clock(self._env)
        self._statistics = StatisticsImpl()

    @property
    def action_store(self) -> ActionStore:
        return self._action_store

    @property
    def exploit_store(self) -> ExploitStore:
        return self._exploit_store

    @property
    def clock(self) -> Clock:
        return self._clock

    @property
    def statistics(self) -> Statistics:
        return self._statistics



from enum import Enum, auto
from typing import Dict, Any, Union, Optional, List, Tuple, Type, Set

from cyst.core.environment.data_store_backend import DataStoreBackend
from cyst.core.environment.data_store_redis_backend import DataStoreRedisBackend
from cyst.core.environment.data_store_memory_backend import DataStoreMemoryBackend

# The data store is quite universal, but its main use-cases are configuration retrieval, log storage and simulation
# artifacts storage


class DataStore:

    def __init__(self, backend_type: str, backend_params: Dict[str, str]) -> None:
        self._backend_type = backend_type.lower()

        self._no_data_store = False
        if self._backend_type == "none":
            self._no_data_store = True
            return

        fn = getattr(self, "configure_" + self._backend_type, self.configure_default)
        self._backend: DataStoreBackend = fn(backend_params) #MYPY: Probably too complex to handle properly? Might be easier to just ignore

    def configure_default(self) -> None:
        raise RuntimeError("Could not find configuration function for backend " + self._backend_type)

    @staticmethod
    def configure_redis(self, host: str, port: int) -> DataStoreBackend:
        backend = DataStoreRedisBackend(host, port)
        if not backend:
            raise RuntimeError("Could not connect client")
        return backend

    @staticmethod
    def configure_memory(self) -> DataStoreBackend:
        return DataStoreMemoryBackend()

    def set(self, run_id: str, item: Any, item_type: Type) -> None:
        if self._no_data_store:
            return

        if not self._backend:
            raise RuntimeError("Data store backend not configured")

        if not isinstance(item, item_type):
            raise RuntimeError("Provided data type {} is not an instance of type {}".format(type(item), item_type))

        self._backend.set(run_id, item, item_type)

    def get(self, run_id: str, item: Any, item_type: Type) -> Any:
        if self._no_data_store:
            return None

        if not self._backend:
            raise RuntimeError("Data store backend not configured")

        return self._backend.get(run_id, item, item_type)

    def update(self, run_id: str, item: Any, item_type: Type) -> None:
        if self._no_data_store:
            return

        if not self._backend:
            raise RuntimeError("Data store backend not configured")

        if not isinstance(item, item_type):
            raise RuntimeError("Provided data type {} is not an instance of type {}".format(type(item), item_type))

        self._backend.update(run_id, item, item_type)

    def remove(self, run_id: str, item: Any, item_type: Type) -> None:
        if self._no_data_store:
            return

        if not self._backend:
            raise RuntimeError("Data store backend not configured")

        self._backend.remove(run_id, item, item_type)

    def clear(self, run_id: str) -> None:
        if self._no_data_store:
            return

        if not self._backend:
            raise RuntimeError("Data store backend not configured")

        self._backend.clear(run_id)

    def commit(self, run_id: str) -> None:
        if self._no_data_store:
            return

        if not self._backend:
            raise RuntimeError("Data store backend not configured")

        self._backend.commit(run_id)


from enum import Enum
from typing import Union, NewType, Any, Type, Dict
from abc import abstractmethod, ABC

from cyst.api.environment.message import Message, MessageType, Response, Request
from cyst.api.environment.stats import Statistics

from cyst.core.environment.message import MessageImpl

read_write_data = [Statistics]
write_only_data = [Message]
append_only_data = [Message]


# ----------------------------------------------------------------------------------------------------------------------
# These functions transform specific interfaces into dicts for storage in data store
def _statistics_to_dict(item: Statistics) -> Dict[str, Any]:
    result: Dict[str, Union[str, float, int]] = dict()

    result["run_id"] = item.run_id
    result["configuration_id"] = item.configuration_id
    result["start_time_real"] = item.start_time_real
    result["end_time_real"] = item.end_time_real
    result["end_time_virtual"] = item.end_time_virtual

    return result


def _message_to_dict(item: Message) -> Dict[str, Any]:
    item: MessageImpl = MessageImpl.cast_from(item)
    result: Dict[str, Union[str, int]] = dict()

    result["type"] = "REQUEST" if item.type == MessageType.REQUEST else "RESPONSE"
    result["id"] = item.id
    result["src_ip"] = str(item.src_ip)
    result["dst_ip"] = str(item.dst_ip)
    result["hop_src_ip"] = str(item.current.ip) #MYPY: IP is not nullable, but current is. Does it make sense to assign None, if current is None, or is it required to fail in some other way to handle such situation? Is this issue even possible
    # and if it is not, are we confidend in that?
    result["hop_src_id"] = item.current.id
    result["hop_dst_ip"] = str(item.next_hop.ip)
    result["hop_dst_id"] = item.next_hop.id
    result["src_service"] = item.src_service
    result["dst_service"] = item.dst_service
    result["ttl"] = item.ttl

    if isinstance(item, Request):
        result["action"] = item.action.id
    else:
        result["action"] = ""

    if isinstance(item, Response):
        result["result"] = "{}|{}|{}".format(item.status.origin.name, item.status.value.name, item.status.detail.name)
    else:
        result["result"] = ""

    return result


def to_dict(item: Union[Statistics, Message]) -> Dict[str, Any]:
    if isinstance(item, Statistics):
        return _statistics_to_dict(item)
    if isinstance(item, Message):
        return _message_to_dict(item)
    else:
        raise RuntimeError("Attempting to convert an item of unknown type '{}' to dictionary.".format(type(item)))


# ----------------------------------------------------------------------------------------------------------------------
class DataStoreBackend:

    @abstractmethod
    def set(self, run_id: str, item: Any, item_type: Type) -> None:
        pass

    @abstractmethod
    def get(self, run_id: str, item: Any, item_type: Type) -> Any:
        pass

    @abstractmethod
    def update(self, run_id: str, item: Any, item_type: Type) -> None:
        pass

    @abstractmethod
    def remove(self, run_id: str, item: Any, item_type: Type) -> None:
        pass

    @abstractmethod
    def clear(self, run_id: str) -> None:
        pass

    @abstractmethod
    def commit(self, run_id: str) -> None:
        pass




from __future__ import annotations

import uuid
import time

from typing import Tuple, TYPE_CHECKING

from cyst.api.environment.control import EnvironmentState, EnvironmentControl
from cyst.api.environment.stats import Statistics
from cyst.api.host.service import ServiceState

from cyst.core.host.service import ServiceImpl, PassiveServiceImpl
from cyst.core.network.node import NodeImpl
from cyst.core.environment.configuration_general import GeneralConfigurationImpl
from cyst.core.environment.serialization import Serializer
from cyst.core.environment.stats import StatisticsImpl

if TYPE_CHECKING:
    from cyst.core.environment.environment import _Environment


class EnvironmentControlImpl(EnvironmentControl):
    def __init__(self, env: _Environment):
        self._env = env

    @property
    def state(self) -> EnvironmentState:
        return _state(self._env)

    def init(self, run_id: str = str(uuid.uuid4())) -> Tuple[bool, EnvironmentState]:
        return _init(self._env, run_id)

    def commit(self) -> None:
        return _commit(self._env)

    def reset(self, run_id: str = str(uuid.uuid4())) -> Tuple[bool, EnvironmentState]:
        return _reset(self._env, run_id)

    def run(self) -> Tuple[bool, EnvironmentState]:
        return _run(self._env)

    def pause(self) -> Tuple[bool, EnvironmentState]:
        return _pause(self._env)

    def terminate(self) -> Tuple[bool, EnvironmentState]:
        return _terminate(self._env)

    def add_pause_on_request(self, id: str) -> None:
        return _add_pause_on_request(self._env, id)

    def remove_pause_on_request(self, id: str) -> None:
        return _remove_pause_on_request(self._env, id)

    def add_pause_on_response(self, id: str) -> None:
        return _add_pause_on_response(self._env, id)

    def remove_pause_on_response(self, id: str) -> None:
        return _remove_pause_on_response(self._env, id)

    def snapshot_save(self) -> str:
        return _snapshot_save(self._env)

    def snapshot_load(self, state: str) -> None:
        return _snapshot_load(self._env, state)

    def transaction_start(self) -> Tuple[int, int, str]:
        return _transaction_start(self._env)

    def transaction_commit(self, transaction_id: int) -> Tuple[bool, str]:
        return _transaction_commit(self._env, transaction_id)

    def transaction_rollback(self, transaction_id: int) -> Tuple[bool, str]:
        return _transaction_rollback(self._env, transaction_id)


# ----------------------------------------------------------------------------------------------------------------------
# Free function implementations of the above class. It is being done this way to shut up the type checking and to
# overcome python's limitation on having a class implemented in multiple files.
def _state(self: _Environment) -> EnvironmentState:
    return self._state


def _init(self: _Environment, run_id: str = str(uuid.uuid4())) -> Tuple[bool, EnvironmentState]:
    if self._initialized:
        return True, self._state

    if self._state == EnvironmentState.RUNNING or self._state == EnvironmentState.PAUSED:
        return False, self._state

    self._pause = False
    self._terminate = False
    self._state = EnvironmentState.INIT
    self._run_id = run_id

    _establish_sessions(self)

    # Set basic statistics
    s = StatisticsImpl.cast_from(self.resources.statistics)
    s.run_id = self._runtime_configuration.run_id if self._runtime_configuration.run_id else self._run_id
    s.configuration_id = self._runtime_configuration.config_id
    s.start_time_real = time.time()

    self._initialized = True

    return True, self._state


def _reset(self: _Environment, run_id: str = str(uuid.uuid4())) -> Tuple[bool, EnvironmentState]:
    if self._state != EnvironmentState.FINISHED and self._state != EnvironmentState.TERMINATED:
        return False, self._state

    self._network.reset()
    self._time = 0
    self._start_time = time.localtime()
    self._tasks.clear()
    self._pause = False
    self._terminate = False
    self._run_id = run_id
    self._state = EnvironmentState.INIT

    return True, self._state


def _establish_sessions(self: _Environment) -> None:
    for session in self._sessions_to_add:
        owner = session[0]
        waypoints = session[1]
        src_service = session[2]
        dst_service = session[3]
        parent = session[4]
        reverse = session[5]

        self.configuration.network.create_session(owner, waypoints, src_service, dst_service, parent, False, reverse)


def _run(self: _Environment) -> Tuple[bool, EnvironmentState]:

    if not self._initialized:
        return False, self._state

    # if paused, unpause
    if self._state == EnvironmentState.PAUSED:
        self._pause = False

    # if this is the first run() after init, call all run() methods of active services and activate passive services
    if self._state == EnvironmentState.INIT:
        for n in GeneralConfigurationImpl.cast_from(self.configuration.general).get_objects_by_type(NodeImpl):
            for s in n.services.values():
                if isinstance(s, ServiceImpl):
                    if s.passive:
                        PassiveServiceImpl.cast_from(s.passive_service).set_state(ServiceState.RUNNING)
                    else:
                        s.active_service.run()

    # Run
    self._state = EnvironmentState.RUNNING
    self._process()

    return True, self._state


def _pause(self: _Environment) -> Tuple[bool, EnvironmentState]:

    if self._state != EnvironmentState.RUNNING:
        return False, self._state

    self._pause = True
    # This will return True + running state, but it will be returned to an actor other than the one who called
    # Environment.run() in the first place. Or I hope so...
    return True, self._state


def _terminate(self: _Environment) -> Tuple[bool, EnvironmentState]:

    if self._state != EnvironmentState.RUNNING:
        return False, self._state

    self._terminate = True
    return True, self._state


def _commit(self: _Environment) -> None:
    s = StatisticsImpl.cast_from(self.resources.statistics)
    s.end_time_real = time.time()
    s.end_time_virtual = self._time

    self._data_store.set(self._run_id, self.resources.statistics, Statistics)


def _add_pause_on_request(self: _Environment, id: str) -> None:
    self._pause_on_request.append(id)


def _remove_pause_on_request(self: _Environment, id: str) -> None:
    self._pause_on_request = [x for x in self._pause_on_request if x != id]


def _add_pause_on_response(self: _Environment, id: str) -> None:
    self._pause_on_response.append(id)


def _remove_pause_on_response(self: _Environment, id: str) -> None:
    self._pause_on_response = [x for x in self._pause_on_response if x != id]


def _snapshot_save(self: _Environment) -> str:
    return Serializer.serialize(self)


def _snapshot_load(self: _Environment, state: str) -> None:
    self._replace(Serializer.deserialize(state))


def _transaction_start(self: _Environment) -> Tuple[int, int, str]:
    return (0, 0, "")


def _transaction_commit(self: _Environment, transaction_id: int) -> Tuple[bool, str]:
    return (True, "")


def _transaction_rollback(self: _Environment, transaction_id: int) -> Tuple[bool, str]:
    return (True, "")



from copy import deepcopy
from typing import List, Optional, Dict, Union, Tuple, Any

from cyst.api.environment.environment import EnvironmentMessaging
from cyst.api.environment.message import Message
from cyst.api.environment.stores import ActionStore, ExploitStore
from cyst.api.environment.resources import EnvironmentResources
from cyst.api.logic.access import AccessLevel
from cyst.api.logic.action import ActionDescription, Action
from cyst.api.logic.exploit import Exploit, ExploitCategory, ExploitLocality
from cyst.api.host.service import PassiveService, ActiveServiceDescription, Service
from cyst.api.network.node import Node

from cyst.core.environment.proxy import EnvironmentProxy
from cyst.core.logic.action import ActionImpl
from cyst.core.host.service import ServiceImpl
from cyst.core.network.session import SessionImpl
from cyst.core.network.node import NodeImpl


class ServiceStoreImpl:

    def __init__(self, messaging: EnvironmentMessaging, resources: EnvironmentResources):
        self._services = {}
        self._messaging = messaging
        self._resources = resources

    def add_service(self, description: ActiveServiceDescription) -> None:
        self._services[description.name] = description

    def get_service(self, name: str) -> Optional[ActiveServiceDescription]:
        if name in self._services:
            return self._services[name]
        else:
            return None

    def create_active_service(self, type: str, owner: str, name: str, node: Node,
                              service_access_level: AccessLevel = AccessLevel.LIMITED,
                              configuration: Optional[Dict[str, Any]] = None, id: str = "") -> Optional[Service]:
        if not type in self._services:
            return None
        node = NodeImpl.cast_from(node)
        proxy = EnvironmentProxy(self._messaging, node.id, name)
        service_description: ActiveServiceDescription = self._services[type]
        service = service_description.creation_fn(proxy, self._resources, configuration)
        return ServiceImpl(type, service, name, owner, service_access_level, id)


class ActionStoreImpl(ActionStore):

    def __init__(self):
        self._actions = {}

    def get(self, id: str = "") -> Optional[Action]:
        if id in self._actions:
            return deepcopy(self._actions[id])
        return None

    def get_ref(self, id: str = "") -> Optional[Action]:
        if id in self._actions:
            return self._actions[id]
        return None

    def get_prefixed(self, prefix: str = "") -> List[Action]:
        result = []
        for id, value in self._actions.items():
            if id.startswith(prefix):
                result.append(deepcopy(value))
        return result

    def add(self, action: ActionDescription) -> None:
        self._actions[action.id] = ActionImpl(action)


class ExploitStoreImpl(ExploitStore):
    def __init__(self):
        # Nested lists here are to allow easier conversion to sets and to do the intersection
        # TODO: revise the data structure logic here
        self._by_id: Dict[str, List[Exploit]] = {}
        self._by_service: Dict[str, List[Exploit]] = {}
        self._by_category: Dict[ExploitCategory, List[Exploit]] = {}

    def clear(self) -> None:
        self._by_id.clear()
        self._by_service.clear()
        self._by_category.clear()

    def add_exploit(self, *exploits: Exploit) -> None:
        for exploit in exploits:
            # Exploit already in store, do nothing
            if exploit.id in self._by_id:
                return

            self._by_id[exploit.id] = [exploit]

            for service in exploit.services.values():
                if service.id not in self._by_service:
                    self._by_service[service.id] = []

                self._by_service[service.id].append(exploit)

            if exploit.category not in self._by_category:
                self._by_category[exploit.category] = []

            self._by_category[exploit.category].append(exploit)

    def get_exploit(self, id: str = "", service: str = "", category: ExploitCategory = ExploitCategory.NONE) -> Optional[List[Exploit]]:
        candidate_sets = []

        if id:
            candidate_sets.append(set(self._by_id.get(id, [])))

        if service:
            candidate_sets.append(set(self._by_service.get(service, [])))

        if category != ExploitCategory.NONE:
            candidate_sets.append(set(self._by_category.get(category, [])))

        if not candidate_sets:
            return []
        else:
            return list(set.intersection(*candidate_sets))

    def evaluate_exploit(self, exploit: Union[str, Exploit], message: Message, node: Node) -> Tuple[bool, str]:
        if isinstance(exploit, str):
            exploits = self.get_exploit(exploit)
            if not exploits:
                return False, "Could not find exploit by id"
            else:
                # Gah!
                exploit = exploits[0]

        # For exploit to be applicable, a number of conditions must be satisfied.
        # 1) Local exploits can only be used at the session end
        if exploit.locality == ExploitLocality.LOCAL:
            if not message.session:
                return False, "Local exploits can only be used from within an existing session"
            elif SessionImpl.cast_from(message.session).endpoint.id != NodeImpl.cast_from(node).id:
                return False, "Local exploits can only be used at session endpoint."

        # 2) Exploits must be applicable on given service_id
        exploit_service = exploit.services.get(message.dst_service, None)
        if not exploit_service:
            return False, "Attempting to use exploit on a service it does not apply to"

        # 3) service_id must exist on node
        node_service = node.services.get(message.dst_service, None)
        if not node_service:
            return False, "The exploit is not fit for any of the services on the node"

        # 4) Versions must match
        if isinstance(node_service, PassiveService) and not exploit_service.min_version <= node_service.version <= exploit_service.max_version:
            return False, "The exploit is not applicable on the current version of a service"

        # 5) TODO There will also be tags to evaluate, but that is a future endeavour

        return True, ""

    @staticmethod
    def cast_from(o: ExploitStore) -> 'ExploitStoreImpl':
        if isinstance(o, ExploitStoreImpl):
            return o
        else:
            raise ValueError("Malformed underlying object passed with the ExploitStore interface")



# The definition here is only to make the type of configuration items obvious.
# Because we are operating on dataclasses, the initialization order precludes us from having some default initialized
# value, which is a real shame (though understandable)
import dataclasses
import uuid
from typing import Any

"""
Provides an API to configure all aspects of the simulation engine.

Available to: creator, models
Hidden from:  agents
"""


class ConfigItem:
    """
    A superclass of a configuration items.
    """
    id: str



from typing import Any, Type

from rejson import Client, Path

from cyst.core.environment.data_store_backend import DataStoreBackend


class DataStoreRedisBackend(DataStoreBackend):

    def __init__(self, host: str, port: int) -> None:
        self._rj = Client(host=host, port=port, decode_responses=True)

    def set(self, run_id: str, item: Any, item_type: Type) -> None:
        pass

    def get(self, run_id: str, item: Any, item_type: Type) -> Any:
        pass

    def update(self, run_id: str, item: Any, item_type: Type) -> None:
        pass

    def remove(self, run_id: str, item: Any, item_type: Type) -> None:
        pass

    def clear(self, run_id: str) -> None:
        pass

    def commit(self, run_id: str) -> None:
        pass



from abc import ABC, abstractmethod
from collections.abc import Sequence
from dataclasses import dataclass, field
from enum import Enum
from flags import Flags
from typing import NamedTuple, List, Tuple, Optional, Any, Dict, Union

from cyst.api.logic.access import AuthenticationToken
from cyst.api.logic.exploit import Exploit


class ActionParameterDomainType(Enum):
    """
    Specifies a type of the action parameter domain.

    Possible values:
        :ANY: The parameters can take any values. If possible avoid this domain, as it does not enable sampling of
            values, thus being less useful for automated processing and learning.
        :RANGE: The parameters belong to a certain numeric range, e.g., port numbers.
        :OPTIONS: The parameters are a collection of values to choose from.
    """
    ANY = 0,
    RANGE = 1,
    OPTIONS = 2


class ActionParameterDomain(Sequence):
    """
    Specifies a domain of action parameters.
    """
    @property
    @abstractmethod
    def type(self) -> ActionParameterDomainType:
        """
        Returns a type of the action parameter domain.

        :rtype: ActionParameterDomainType
        """

    @property
    @abstractmethod
    def range_min(self) -> int:
        """
        Returns a minimal value of the range, if the domain is of RANGE type. Otherwise the value is undefined.

        :rtype: int
        """

    @property
    @abstractmethod
    def range_max(self) -> int:
        """
        Returns a maximal value of the range, if the domain is of RANGE type. Otherwise the value is undefined.

        :rtype: int
        """

    @property
    @abstractmethod
    def range_step(self) -> int:
        """
        Returns a gap between two values in a range. E.g., a range [1,3,5,7] has a step = 2.

        :rtype: int
        """

    @property
    @abstractmethod
    def options(self) -> List[Any]:
        """
        Returns a list of possible values in the domain, if the type of the domain is OPTIONS. Otherwise and empty list.

        :rtype: List[Any]
        """

    @abstractmethod
    def validate(self, value: Any) -> bool:
        """
        Checks, whether a value belongs into a domain.

        :param value: A value to check.
        :type value: Any
        """

    @property
    @abstractmethod
    def default(self) -> Any:
        """
        :return: A value from the domain that is considered a default one.
        """

    @abstractmethod
    def __getitem__(self, item: int) -> Any:
        """
        This method is here for the domain to be treated as a range, so it can be, e.g., randomly sampled. It is not
        intended to be used directly. This does not work with domains of type ANY.

        :param item: An index of the item to get.
        :type item: int

        :return: The item with given index.
        """

    @abstractmethod
    def __len__(self) -> int:
        """
        This method is here for the domain to be treated as a range, so it can be, e.g., randomly sampled. It is not
        intended to be used directly. This does not work with domains of type ANY.

        :return: The number of items in the domain.
        """


class ActionParameterType(Enum):
    """
    Specifies a type of action parameters.

    Possible values:
        :NONE: Type not specified, or outside the current domain of parameter types. This should be used only after
            due consideration, as it is better to expand this domain than to use NONE. It may interfere with the
            learning process.
        :IDENTITY: The parameter is an identity of a user in the system.
        :IDENTIFIER: The parameter is some kind of identifier.
        :DURATION: The parameter represents a time duration in simulated time units.
        :TOKEN: The parameter is an instance of a token object, such as cyst.api.logic.access.AuthenticationToken.
        :DATA: Added as a temporary solution for actions and communication that carries actual data not encoded into
            the message structure. Given the complicated definition of parameters and factual uselessness of fitting
            data into a domain that can be sampled, this will likely be replaced by a more top-level solution.
    """
    NONE = 0,
    IDENTITY = 1,
    IDENTIFIER = 2,
    DURATION = 3,
    TOKEN = 4,
    DATA = 5

@dataclass
class ActionParameter:
    """
    Action parameter represents a mechanism to further specify details of actions. An action can have an arbitrary
    number of parameters.

    :param type: The type of the parameter.
    :type type: ActionParameterType

    :param name: A name of the parameter, which is unique within an action.
    :type name: str

    :param domain: A domain of the parameter.
    :type domain: ActionParameterDomain

    :param value: A default value of the parameter.
    :type value: Optional[Any]
    """
    type: ActionParameterType
    name: str
    domain: ActionParameterDomain
    value: Optional[Any] = None


class ActionToken(Flags):
    """
    Action tokens are used to describe at the high level, which tokens are required for the action to properly execute
    in the context of an attack graph. Tokens are primarily used for automatic generation of simulation scenarios.
    They are also currently being restructured.

    An example usage:

        .. code-block::

            f1 = ActionToken.NONE
            f2 = ActionToken.DATA | ActionToken.SESSION

    Possible values:
        :NONE: No additional token is required. An example is a scanning action.
        :AUTH: Correct authorization or authentication tokens are required.
        :DATA: Specific data must be obtained or produced.
        :EXPLOIT: A working exploit must be present for this action to work.
        :SESSION: A session must be established to the node the action is being executed on.
    """
    NONE = (),
    AUTH = (),
    DATA = (),
    EXPLOIT = (),
    SESSION = ()


@dataclass
class ActionDescription:
    """
    A description of an action in the form that is used for registering into the system.

    :param id: A unique name of the action. This name must be unique across all behavioral models, so it is advisible
        to use own namespace for each model.
    :type id: str

    :param description: A textual description of the action.
    :type description: str

    :param parameters: A list of action parameters.
    :type parameters: List[ActionParameter]

    :param tokens: A list of action token tuples. Each tuple represents inputs and outputs like this (input tokens,
        output tokens). Because a token is a flag, multiple tokens can be supplied by means of the OR operator, e.g.:

        .. code-block::

            (ActionToken.AUTH | ActionToken.EXPLOIT, ActionToken.SESSION).
    """
    id: str
    description: str
    parameters: List[ActionParameter]
    tokens: List[Tuple[ActionToken, ActionToken]]


class Action(ABC):
    """
    An action represent an activity that an actor within simulation can do. Usage of actions is mostly the domain of
    active services. An action does not carry any semantics by itself. The semantics is provided by interpreters /
    behavioral models. One action can therefore have different impact depending on the model chosen.
    """
    @property
    @abstractmethod
    def id(self) -> str:
        """
        A unique identifier of the action within simulation. The identifier is expected to be a serialization of
        a position in hierarchy, separated by a colon. E.g., "aif:privilege_escalation:root_privilege_escalation".

        :rtype: str
        """

    @property
    @abstractmethod
    def description(self) -> str:
        """
        A textual description of the action.

        :rtype: str
        """

    @property
    @abstractmethod
    def fragments(self) -> List[str]:
        """
        The list of fragments of the id (i.e., what is left after removing the colons). E.g, ["aif",
        "privilege_escalation", "root_privilege escalation"].

        :rtype: List[str]
        """

    @property
    @abstractmethod
    def namespace(self) -> str:
        """
        Currently a shorthand for fragments()[0], but that can change in the future, as the namespace can span multiple
        fragments.

        :rtype: str
        """

    @property
    @abstractmethod
    def exploit(self) -> Optional[Exploit]:
        """
        Returns an exploit currently associated with the action.

        :rtype: Optional[Exploit]
        """

    @abstractmethod
    def set_exploit(self, exploit: Optional[Exploit]) -> None:
        """
        Sets an exploit associated with the action. TODO: Why the hell is this not a setter of the exploit property?

        :param exploit: The exploit to use. If None is passed as a value, removes the exploit from the action.
        :type exploit: Optional[Exploit]
        """

    @property
    @abstractmethod
    def parameters(self) -> Dict[str, ActionParameter]:
        """
        Returns a collection of action parameters.

        :rtype: Dict[str, ActionParameter]
        """

    @abstractmethod
    def add_parameters(self, *params: ActionParameter) -> None:
        """
        Adds one or more parameters to the action. If a parameter of the same name is present, it gets overwritten.

        :param params: One or more action parameters.
        :type params: ActionParameter
        """

    @property
    @abstractmethod
    def tokens(self) -> List[Tuple[ActionToken, ActionToken]]:
        """
        Returns a list of action tokens, which define inputs and outputs of the action.

        :rtype: List[Tuple[ActionToken, ActionToken]]
        """

    @abstractmethod
    def copy(self) -> 'Action':
        """
        Returns a copy of the action. This is useful for keeping copies of action with different parameters or exploits.

        :rtype: Action
        """



from setuptools import setup, find_packages, find_namespace_packages
import pathlib

here = pathlib.Path(__file__).parent.resolve()

# Get the long description from the README file
long_description = (here / 'README.md').read_text(encoding='utf-8')

# Arguments marked as "Required" below must be included for upload to PyPI.
# Fields marked as "Optional" may be commented out.

setup(
    name='cyst-core',
    version='0.5.0',
    description='API and runtime of the CYST framework',
    long_description=long_description,
    long_description_content_type='text/markdown',
    url='https://muni.cz/go/cyst/',
    author='Martin Drasar et al.',
    author_email='drasar@ics.muni.cz',
    classifiers=[
        # How mature is this project? Common values are
        #   3 - Alpha
        #   4 - Beta
        #   5 - Production/Stable
        'Development Status :: 4 - Beta',

        # Indicate who your project is intended for
        'Intended Audience :: Developers',
        'Intended Audience :: Science/Research',
        'Topic :: Scientific/Engineering',
        'Topic :: Scientific/Engineering :: Artificial Intelligence',
        'Topic :: Security',
        'Typing :: Typed',

        # Pick your license as you wish
        'License :: OSI Approved :: GNU Lesser General Public License v3 (LGPLv3)',

        # Specify the Python versions you support here. In particular, ensure
        # that you indicate you support Python 3. These classifiers are *not*
        # checked by 'pip install'. See instead 'python_requires' below.
        'Programming Language :: Python :: 3',

        'Operating System :: OS Independent'
    ],
    packages=find_packages(exclude=['tests', 'docs']) + find_namespace_packages(include=['cyst_models.*', 'cyst_services.*']),
    python_requires='>=3.9, <4',

    # This field lists other packages that your project depends on to run.
    # Any package you put here will be installed by pip when your project is
    # installed, so they must be valid existing projects.
    #
    # For an analysis of "install_requires" vs pip's requirements files see:
    # https://packaging.python.org/discussions/install-requires-vs-requirements/
    install_requires=[
        'netaddr',
        'py-flags',
        'semver',
        'rejson',
        'networkx',
        'cachetools',
        'pyserde',
        'importlib_metadata',
        'jsonpickle'
    ],

    # To provide executable scripts, use entry points in preference to the
    # "scripts" keyword. Entry points provide cross-platform support and allow
    # `pip` to create the appropriate form of executable for the target
    # platform.
    #
    # For example, the following would provide a command called `sample` which
    # executes the function `main` from this package when invoked:
    entry_points={
        'cyst.models': [
            'cyst=cyst_models.cyst.main:action_interpreter_description',
            'meta=cyst_models.meta.main:action_interpreter_description'
        ],
        'cyst.services': [
            'scripted_actor=cyst_services.scripted_actor.main:service_description',
            'forward_shell=cyst_services.forward_shell.main:service_description',
            'reverse_shell=cyst_services.reverse_shell.main:service_description'
        ]
    },

    # List additional URLs that are relevant to your project as a dict.
    #
    # This field corresponds to the "Project-URL" metadata fields:
    # https://packaging.python.org/specifications/core-metadata/#project-url-multiple-use
    #
    # Examples listed include a pattern for specifying where the package tracks
    # issues, where the source is hosted, where to say thanks to the package
    # maintainers, and where to support the project financially. The key is
    # what's used to render the link text on PyPI.
    #project_urls={  # Optional
    #    'Bug Reports': 'https://github.com/pypa/sampleproject/issues',
    #    'Funding': 'https://donate.pypi.org',
    #    'Say Thanks!': 'http://saythanks.io/to/example',
    #    'Source': 'https://github.com/pypa/sampleproject/',
    #},
)



from abc import ABC, abstractmethod
from time import struct_time
from typing import Any

from cyst.api.host.service import ActiveService


class Clock(ABC):
    """ Clock interface provides an access to reading the simulation time. It can also be used to set a timeout for a
    service.
    """
    @abstractmethod
    def simulation_time(self) -> int:
        """ Returns a current time of simulation in internal simulation units.

        :return: Simulation time.
        """

    @abstractmethod
    def hybrid_time(self) -> struct_time:
        """ Returns a current time of simulation in pseudo-real time, which is calculated as a base time, given during
        initialization + (clock factor * simulation time).

        :return: Simulation time.
        """

    @abstractmethod
    def timeout(self, service: ActiveService, delay: int, content: Any) -> None:
        """ Schedule a timeout message for a given service. This acts like a time callback and enables inclusion of
        any kind of data.

        :param service: The service, which should receive the timeout message.
        :param delay: The duration of the timeout in simulation time.
        :param content: The included data. They will not be modified.
        :return: None
        """



from typing import Union, Optional, Any
from netaddr import IPAddress

from cyst.api.environment.environment import EnvironmentMessaging
from cyst.api.environment.messaging import Message, Request, Status
from cyst.api.logic.access import Authorization
from cyst.api.logic.action import Action
from cyst.api.network.session import Session

from cyst.core.environment.message import RequestImpl
from cyst.core.network.elements import Endpoint


# EnvironmentProxy is a proxy for the environment, which is passed to each active node. It takes care of routing of
# messages and prevents forging of Messages and spooky action in the distance
class EnvironmentProxy(EnvironmentMessaging):

    def __init__(self, env: EnvironmentMessaging, node_id: str, service_id: str) -> None:
        self._env = env
        # Node is resolved on the first attempt to send a message
        self._node_id = node_id
        self._service_id = service_id

    def send_message(self, message: Message, delay: int = 0) -> None:
        # Dummy origin, to make it work with Environment.send_message
        if isinstance(message, RequestImpl):
            message.set_origin(Endpoint(self._node_id, -1))
            message.src_service = self._service_id

        self._env.send_message(message, delay)

    def create_request(self, dst_ip: Union[str, IPAddress], dst_service: str = "",
                       action: Action = None, session: Session = None, auth: Authorization = None) -> Request:
        return self._env.create_request(dst_ip, dst_service, action, session, auth)

    def create_response(self, request: Request, status: Status, content: Optional[Any] = None,
                        session: Optional[Session] = None, auth: Optional[Authorization] = None):
        return self._env.create_response(request, status, content, session, auth)

    def open_session(self, request: Request) -> Session:
        return self._env.open_session(request)
